import os, sys
import torch.onnx
import functools
#no complete
NORM_LAYER_KEYWORDS = ["BatchNormalization","batch_norm", "group_norm", "instance_norm"]
CONV_LAYER_KEYWORDS = ["convolution", "addmm", "Conv"]  # does not support groups > 1 for conv
LINEAR_LAYER_KEYWORDS = ["Gemm",""] # "PRelu"
#PASS_KEYWORDS = ["Mul","relu", "leaky_relu", "sigmoid", "tanh","Relu",
#                 "pool", "pad", "dropout","MaxPool","Reshape","Concat","Unsqueeze","Gather","Shape","GlobalAveragePool",
#                 "view", ]  # and more .. does not support concat
PASS_KEYWORDS = ["Mul","relu", "PRelu", "Prelu","leaky_relu", "sigmoid", "tanh","Relu",
                 "pool", "pad", "Pad", "dropout","MaxPool","Reshape","Unsqueeze","Gather","Shape","GlobalAveragePool",
                 "view", "Resize","Dropout" ,"dropout","AveragePool"]  # and more .. does not support concat
ADD_KEYWORDS = ["Add" ]  #
UN_SUPPORTED_KEYWORDS = ["Concat", ]
#OTHER_OP_KEYWORDS = ["cat"]
#OTHER_PRIM_KEYWORDS = ["ListConstruct"]  # for cat
#NO_EFFECT_KEYWORDS = ["size", ]

#def process_prune_layers(opt):
 #   opt.model

def get_prune_layers(opt):
    """load pruned weights to a unpruned model instance

    Arguments:
        net (pytorch model): the cpu model instance
        opt (easypruner.): pruned weights
    step:
         1. get onnx file
         2. get graph, succ&prec relation by onnx file
         3. merge the passable layer
         3. get norm BN
         4. get Add operation 
         5. get connectate  operation
    Returns:
        None
        the result return by
        opt.norm_layer_names 
        opt.prec_layers     
        opt.succ_layers

        a model instance with pruned structure (and weights if load_pruned_weights==True)
    """
    net = opt.model 
    to_del = []
    def GetPydotGraph(
    graph,  # type: GraphProto
    name=None,  # type: Optional[Text]
    rankdir='LR',  # type: Text
    node_producer=None,  # type: Optional[_NodeProducer]
    embed_docstring=False,  # type: bool
    ):  # type: (...) -> pydot.Dot

        previous_layer = {}
        succeding_layer = {}
        blob_to_layer={}
       # import pdb;pdb.set_trace()
        for op_id, op in enumerate(graph.node):
            name = op_id;#op.name
            for output_name in op.output:
                if output_name in blob_to_layer:
                    print("error");
#                    import pdb;pdb.set_trace()
                blob_to_layer[output_name] = name;

            for input_name in op.input:
                if input_name not in blob_to_layer:
                    input_node = None
                    blob_to_layer[input_name] = None
                    #import pdb;pdb.set_trace()
                else:
                    input_layer = blob_to_layer[input_name]
                    if name in previous_layer:
                        previous_layer[name].append(input_layer)
                    else:
                        previous_layer[name] = [input_layer,]
                    if input_layer in succeding_layer:
                        succeding_layer[input_layer].append(name)
                    else:
                        succeding_layer[input_layer] = [name,]
        return previous_layer,succeding_layer

    def omit_pass(node_id,biao,org_dict, new_dict):
        if biao not in org_dict:
            return
        biao = org_dict[biao]
        for i in biao:
            if model.graph.node[i].op_type in PASS_KEYWORDS :
                omit_pass( node_id, i,org_dict,new_dict)
            elif model.graph.node[i].op_type in UN_SUPPORTED_KEYWORDS:
                #new_dict[node_id].append(-1000)
                to_del.append(node_id)
            else :
                if i not in new_dict[ node_id ]:
                    new_dict[ node_id ].append(i)
    def omit_pass_forward(node_id,biao,org_dict, new_dict):
        if biao not in org_dict:
            return
        biao = org_dict[biao]
        for i in biao:
            if model.graph.node[i].op_type in  ['shape','Shape'] :
                continue
            if model.graph.node[i].op_type in PASS_KEYWORDS :
                omit_pass_forward( node_id, i,org_dict,new_dict)
            elif model.graph.node[i].op_type in UN_SUPPORTED_KEYWORDS:
                to_del.append(node_id)
            #    new_dict[node_id].append(-1000)
            else :
                if i not in new_dict[ node_id ]:
                    new_dict[ node_id ].append(i)
    #1. get onnx file
    #import pdb;pdb.set_trace()
    if opt.onnx_file == None:
        if type(net.forward) is functools.partial: #for mmlab #opt.is_open_mmlab_codes:
            torch.onnx.export( net.cpu(), [torch.randn(1, *opt.input_shape, requires_grad=True)],  os.path.join( opt.workdir, opt.exp_name+'.onnx') ,verbose= False, opset_version=11, training= True, do_constant_folding=True)
        else:
            torch.onnx.export( net.cpu(), torch.randn(1, *opt.input_shape, requires_grad=True),  os.path.join( opt.workdir, opt.exp_name+'.onnx') ,verbose= False, opset_version=11, training= True, do_constant_folding=True)
        import onnx 
        model = onnx.load(os.path.join( opt.workdir, opt.exp_name+'.onnx'))
    else:
        import onnx 
        model = onnx.load(opt.onnx_file)
        print("load my onnx file")
        

    #2. get graph, succ&prec relation
    previous_layer,succeding_layer = GetPydotGraph(model.graph)

    ####################################################### get the layer name
    BN_layer = []
    for op_id, op in enumerate(model.graph.node):
        if op.op_type in NORM_LAYER_KEYWORDS:
            BN_layer.append(op_id)

    ADD_layer = []
    for op_id, op in enumerate(model.graph.node):
        if op.op_type in ADD_KEYWORDS:
            ADD_layer.append(op_id)

    CONV_layer = []
    for op_id, op in enumerate(model.graph.node):
        if op.op_type in CONV_LAYER_KEYWORDS:
            CONV_layer.append(op_id)
    ####################################################### merge the passable layer
#    import pdb;pdb.set_trace()
    prec_l = {}
    succ_l = {}
    for i in BN_layer + ADD_layer : #+ CONV_layer :
        prec_l[i] = []
        succ_l[i] = []
        omit_pass_forward(i, i, succeding_layer, succ_l)
        omit_pass(i, i, previous_layer,  prec_l)

#    to_del=[]
#    for key  in succ_l:
#        for j in succ_l[key]:
#            if j == -1000:
#                #del succ_l[key]
#                to_del.append(key)
#                #break
#    for key in to_del:
#        if key in succ_l:
#            del succ_l[key]
#        if key in prec_l:
#            del prec_l[key]
#        if key in BN_layer:
#            del BN_layer[key]
 
            
#    import pdb;pdb.set_trace()
    result_BN = []
    result_succ = []
    result_prec = []

    BN2add_layer =[]
    ######################################################## batch normal
    for i in BN_layer :
        if "Add" in [  model.graph.node[j].op_type   for j in succ_l[i]    ]:
            BN2add_layer.append(i)
            continue
        result_BN.append([i])
        result_succ.append(succ_l[i])
        result_prec.append(prec_l[i])

    ############################################################add
    add_bn_layer = []
    add_bn_succ = []
    add_bn_prec = []

    def add_get_root(node_id):
        if not( set(ADD_KEYWORDS)  & set(  model.graph.node[i].op_type   for i in prec_l[node_id]  )):
            return node_id
        for i in prec_l[node_id] :
            if model.graph.node[i].op_type in ADD_KEYWORDS:
                return add_get_root(i)

    def add_get_all(node_id,deal_list):
        deal_list.append(node_id)
        for i in succ_l[node_id]:
            if model.graph.node[i].op_type in ADD_KEYWORDS:
                add_get_all(i,deal_list)

    have_deal = []
    for i in ADD_layer:
        #judge
        if i in have_deal:
            continue
        #get all add layer
        Add_root = add_get_root(i)
        deal_list = []
        #print("root",Add_root)
        add_get_all(Add_root,deal_list)
        #print(deal_list)
        #get_bn_layer
        bn_layer = []
        for ii in deal_list:
            for j in prec_l[ii]:
                if model.graph.node[j].op_type in   NORM_LAYER_KEYWORDS :
                    bn_layer.append(j)
        #get_prec
        add_prec = []
        for ii in bn_layer:
            for j in prec_l[ii]:
                add_prec.append(j)


        #get_succ
        add_succ =set()
        for ii in bn_layer + deal_list:
            for j in succ_l[ii]:
                if model.graph.node[j].op_type not in  ADD_KEYWORDS:
                    add_succ.add(j)
       
        #dealed
        have_deal += deal_list
        add_bn_layer.append( bn_layer   )
        add_bn_succ.append(  list(add_succ)  )
        add_bn_prec.append(  add_prec  )




    #########################################################################deal all
    if opt.add_layer_need :
        result_BN   += add_bn_layer
        result_prec += add_bn_prec
        result_succ += add_bn_succ
    #import pdb;pdb.set_trace();
    """
    if opt.stride_2:
        for lid,i in enumerate(result_prec):
            for ele in i:
                node = model.graph.node[ele]
                for attribute in node.attribute:
                    if attribute.name == "strides":
                        strides = max(attribute.ints)
                        if strides > 1:
                            for bnele in result_BN[lid]:
                                to_del.append(bnele)
                            continue
    """
  
   # import pdb;pdb.set_trace()
    to_save_indexes = []
    for lid, layer in enumerate(result_BN):
        delete_flag = False;
        for node_id in layer:
            if node_id in to_del:
                delete_flag = True;
        if delete_flag:
            pass
        else:
            to_save_indexes.append(lid)


  #  import pdb;pdb.set_trace();
    result_BN = [result_BN[ele] for ele in to_save_indexes]
    result_prec = [result_prec[ele] for ele in to_save_indexes]
    result_succ = [result_succ[ele] for ele in to_save_indexes]

    # for to_del_id in to_del_indexes:
    #     del result_BN[to_del_id] 
    #     del result_prec[to_del_id]
        
        
    #     del result_succ[to_del_id]
    #import pdb;pdb.set_trace();

    result_BN_str     = []
    result_prec_str   = []
    result_succ_str   = []
    for i in range(len(result_BN)):
        list_=[]
        for j in result_BN[i]:
            weight_name = model.graph.node[j].input[1]
            group_name = '.'.join(weight_name.split('.')[:-1])
            list_.append(group_name)
            if group_name=="":
                #import pdb;pdb.set_trace()
                pass;
        result_BN_str.append(list_)
        
        list_=[]
        for j in result_prec[i]:
            if len( model.graph.node[j].input ) > 1:
                weight_name = model.graph.node[j].input[1]
                group_name = '.'.join(weight_name.split('.')[:-1])
                list_.append(group_name)
                if group_name=="":
                    #import pdb;pdb.set_trace()
                    pass;
        result_prec_str.append(list_)
        
        list_=[]
        for j in result_succ[i]:
            if len( model.graph.node[j].input ) > 1:
                #import pdb;pdb.set_trace()
                weight_name = model.graph.node[j].input[1]
                group_name = '.'.join(weight_name.split('.')[:-1])
                list_.append(group_name)
                if group_name=="":
                    #import pdb;pdb.set_trace()
                    pass;
        result_succ_str.append(list_)
#    import pdb;pdb.set_trace();
    opt.norm_layer_names =  result_BN_str
    opt.prec_layers      =  result_prec_str
    opt.succ_layers      =  result_succ_str  
    print()
    #import pdb;pdb.set_trace();



from datetime import datetime
from . import   config



def fastpruner(net,prune_factor=0.5,input_dim = [3,416,416],method = 'Ratio', onnx_file=None):
    if method == 'Uniform':
        fastuniform(net,prune_factor ,input_dim,onnx_file)
    elif method == 'Ratio':
        fastratio(net,prune_factor,input_dim,onnx_file)
    elif method == 'Order':
        fastorder(net,prune_factor,input_dim,onnx_file)
    else:
        print("please check the format of methods.[Uniform, Ratio, order]")


def fastuniform(net,flops_saved=0.5,input_dim = [3,224,224],onnx_file=None):
    opt = config('tmp','Uniform' )
    opt.add_layer_need = False
    opt.concat_layer_need = False
    opt.model = net
    opt.onnx_file = onnx_file
    opt.use_conv_statistics = True
    #opt.prune_stride_2 = False
    opt.input_shape= (3 ,  input_dim[1], input_dim[2]  )     #input_shape   (3,256,256)
    #prune
    opt.prune_save_flops  = flops_saved    #save_flops float
    opt.prune_criterion   = 'L1'    #criterion()   importance函数
    opt.init()
    opt.prune_algorithm_onlyprune()

def fastratio(net,channel_saved=0.5,input_dim = [3,224,224],onnx_file=None):
    opt = config('tmp','Ratio' )
    opt.algorithm = 'Ratio'
    opt.add_layer_need = False
    opt.concat_layer_need = False
    opt.model = net
    opt.input_shape = (3 ,  input_dim[1], input_dim[2] )     #input_shape   (3,256,256)
    #prune
    opt.prune_criterion   = 'L1'    #criterion()   importance函数
    opt.prune_factor =channel_saved
    opt.onnx_file = onnx_file
    opt.use_conv_statistics = True
    opt.prune_stride_2  = False
    opt.init()
    opt.prune_algorithm_onlyprune()



def fastorder(net, threshold=0.5,input_dim = [3,224,224],onnx_file=None):
    opt = config('tmp','Order' )
    opt.algorithm = 'Order'
    opt.add_layer_need = False
    opt.concat_layer_need = False
    opt.model = net
    opt.onnx_file = onnx_file
    opt.use_conv_statistics = False
    opt.prune_stride_2  = True
    opt.input_shape = (3 ,  input_dim[1], input_dim[2] )     #input_shape   (3,256,256)
    #prune
    opt.prune_criterion   = 'L1'    #criterion()   importance函数
    opt.prune_factor = threshold
    opt.init()
    opt.prune_algorithm_onlyprune()



def getprunelayer(net,flops_saved=0,input_dim = [3,224,224],onnx_file=None):
    opt = config('tmp','tmp' )
    opt.add_layer_need = False
    opt.concat_layer_need = False
    opt.prune_stride_2 = True
    opt.model = net
    opt.onnx_file = onnx_file
    opt.input_shape= (3 ,  input_dim[1], input_dim[2]  )     #input_shape   (3,256,256)
    opt.prune_save_flops  = flops_saved    #save_flops float
    opt.prune_criterion   = 'L1'    #criterion()   importance函数
    opt.init()
    return  opt.sparsity_layernames





#import os 
#import core 
#import  prune 
#import regularize
#import scheme
#import utils 

from ..prune  import  prune
__all__ = ['prune_finetune' , 'sparsity_prune_finetune' , 'prune_op' ]

def prune_finetune(opt):
    prune(opt)
    if opt.Adaptive_BN is not None and opt.dataloader_train is not None:
        opt.Adaptive_BN(opt.model , opt.dataloader_train)
        opt.eval(opt.model , opt.args)
    opt.args.lr = 1e-3
    opt.args.state =  'finetune'
    opt.args.path = opt.workdir
    opt.train(opt.model , opt.args)
    
def sparsity_prune_finetune(opt):

    opt.args.state = 'sparsity'
    opt.args.EP_norm_layer_names = opt.sparsity_layernames
    opt.args.EP_sparsity_factor= opt.sparsity_factor  
    opt.args.EP_mask_dict = opt.sparsity_mask
    opt.args.EP_regular_method = opt.sparsity_method

    opt.train(opt.model , opt.args)
    prune(opt)
    opt.args.lr = 1e-3 

    opt.args.state =  'finetune'
    opt.train(opt.model , opt.args) 

def prune_op(opt):
    prune(opt)
    if opt.Adaptive_BN is not None and opt.dataloader_train is not None:
        opt.Adaptive_BN(opt.model , opt.dataloader_train)






#from .arch import DDPG , EagleEye, Uniform, sensity_analysis
from .arch import Uniform , EagleEye
from .threshold import  get_mask_threshold
from ..utils.flops_counter import get_model_complexity_info
from ..utils.judge import is_transposed

from .utils import get_pruning_mask ,pruning_model_by_mask
import torch
import torch.nn as nn
from ..utils.rebuild import rebuild
import os 

__all__ = ['prune' ]

MIN_SCALING_FACTOR = 1e-18
OT_DISCARD_PERCENT = 0.0001
OUT_CHANNEL_DIM = 0
IN_CHANNEL_DIM = 1
WEIGHT_POSTFIX = ".weight"
BIAS_POSTFIX = ".bias"
MIN_CHANNELS = 1


def get_channel_numbers_arch(opt):
    """Get the Architecture of the pruned model by the method in the target flops.
    
    Args:
        model: the torch net to be pruned.
        method(str): the method of getting channel_config 
                ['DDPG' , 'EAgleEye', 'Uniform', 'seneity_analysis']
                1. sensity analysis method 
                2. uniform pruning
                3. DDPG
                4. EagleEye
        flops_ratio(float):  the flops target of the pruned model.
    Returns:
        channel_number_dict(dict): A dict of chennel numbers of every layer. Like { 'layer1.bn':2 , 'layer1.bn1':7 ,.....}
    """
    #定义agent
    #if method =='DDPG':
    #    agent = DDPG(opt)
    if opt.prune_method =='EagleEye':
        agent = EagleEye(opt)
    elif opt.prune_method =='Uniform':
        agent = Uniform(opt)
    #elif method =='Sensity_analysis':
    #    agent = Sensity_analysis(opt)
    agent.run()
   
    return agent.result

def  get_channel_numbers_threshold(opt):
    """Get the Architecture of the pruned model by the method in the threshold.
    
    Args:
        model: the torch net to be pruned.
        method(str): the method of getting channel_config 
                ['Order' , 'Ratio', 'OT']
        flops_ratio(float):  the flops target of the pruned model.
    Returns:
        channel_number_dict(dict): A dict of chennel numbers of every layer. Like { 'layer1.bn':2 , 'layer1.bn1':7 ,.....}
    """
    if opt.use_conv_statistics:
        pruned_names = opt.prec_layers
    else:
        pruned_names =  opt.norm_layer_names

    assert opt.prune_factor is not None, 'prune_factor is None!'
    channel_savenumber_dict_ = get_mask_threshold(opt.model, prune_factor = opt.prune_factor, criterion = opt.prune_criterion, pruned_names = pruned_names, threshold_method = opt.prune_method)
    
    channel_savenumber_dict = {}
    if opt.use_conv_statistics:
        for prec_name, norm_name in zip( opt.prec_layers , opt.norm_layer_names ):
            channel_savenumber_dict[ norm_name[0]   ] = channel_savenumber_dict_[ prec_name[0]  ]
    else:
        channel_savenumber_dict = channel_savenumber_dict_ 
    return  channel_savenumber_dict

def get_mask_model(opt):#接口函数
    """Get the pruning mask by the order algorithm(option).
    Args:
        model: the torch net to be pruned.
        opt: pruner config dict
    Returns:
        pruning_mask(dict): the pruning mask ,Like { 'layer1.bn':torch.Tensor([True,False,False,True,...]) , ....}
    """
    assert opt.prune_method in ['DDPG' , 'EagleEye', 'Uniform', 'Seneity_analysis', 'Order', 'Ratio', 'OT'], opt.prune_method + ' is not in  [\'DDPG\' , \'EagleEye\', \'Uniform\', \'seneity_analysis\', \'order\', \'ratio\', \'OT\'] '

    if opt.prune_method in ['DDPG' , 'EagleEye', 'Uniform', 'Seneity_analysis']:
        channel_savenumber_dict = get_channel_numbers_arch(opt)                    
    elif opt.prune_method in ['Order', 'Ratio', 'OT']:
        channel_savenumber_dict =  get_channel_numbers_threshold(opt)

    pruning_mask = get_pruning_mask(opt , channel_savenumber_dict )
    return pruning_mask
 

def prune( opt):
    #获取剪枝mask
    if opt.prune_mask is None:
        opt.prune_mask = get_mask_model(opt)

    #mask剪枝
    opt.model =  pruning_model_by_mask(  opt.model , opt.prune_mask  ,  opt.norm_layer_names,  opt.prec_layers,  opt.succ_layers)
    
    #update pruning state
    import copy
    opt.pruned = False
    opt.prunedmodel_flops ,opt.prunedmodel_params = get_model_complexity_info( copy.deepcopy(opt.model) , ( opt.input_shape ) , print_per_layer_stat=False )
    opt.real_prune_save_flops = opt.prunedmodel_flops /opt.flops
    
    #保存mask
    torch.save(opt.prune_mask , os.path.join( opt.workdir, opt.exp_name+'_'+ ( "%.2f" % opt.real_prune_save_flops) +'_flops.mask'))
    #保存权重
    torch.save(opt.model.state_dict() , os.path.join( opt.workdir, opt.exp_name+'_'+ ( "%.2f" % opt.real_prune_save_flops) +'_flops.pth'))





# train.py
#!/usr/bin/env	python3

""" train network using pytorch

author baiyu
"""

import os
import sys
import argparse
from datetime import datetime

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
import torchvision.transforms as transforms

from torch.utils.data import DataLoader
#from dataset import *
from torch.autograd import Variable

#from tensorboardX import SummaryWriter
import numpy as np
from .judge import is_number , get_module

MIN_SCALING_FACTOR = 1e-18
OT_DISCARD_PERCENT = 0.0001
OUT_CHANNEL_DIM = 0
IN_CHANNEL_DIM = 1
WEIGHT_POSTFIX = ".weight"
BIAS_POSTFIX = ".bias"
MIN_CHANNELS = 1


def load_pruned_model(model, pruned_weights, prefix='', load_pruned_weights=True, inplace=True):
    """load pruned weights to a unpruned model instance

    Arguments:
        model (pytorch model): the model instance
        pruned_weights (OrderedDict): pruned weights
        prefix (string optional): prefix (if has) of pruned weights
        load_pruned_weights (bool optional): load pruned weights to model according to the ICLR 2019 paper:
            "Rethinking the Value of Network Pruning", without finetuning, the model may achieve comparable or even
            better results
        inplace (bool, optional): if return a copy of the model

    Returns:
        a model instance with pruned structure (and weights if load_pruned_weights==True)
    """
    
    def _dirty_fix(module, param_name, pruned_shape):
        module_param = getattr(module, param_name)
        # identify the dimension to prune
        pruned_dim = 0
        for original_size, pruned_size in zip(module_param.shape, pruned_shape):
            if original_size != pruned_size:
                keep_indices = torch.LongTensor(range(pruned_size)).to(module_param.data.device)
                module_param.data = module_param.data.index_select(pruned_dim, keep_indices)
  
                # modify number of features/channels
                if param_name == "weight":
                    if isinstance(module, nn.modules.batchnorm._BatchNorm) or \
                            isinstance(module, nn.modules.instancenorm._InstanceNorm) or \
                            isinstance(module, nn.GroupNorm):
                        module.num_features = pruned_size
                    elif isinstance(module, nn.modules.conv._ConvNd):
                        if pruned_dim == OUT_CHANNEL_DIM:
                            module.out_channels = pruned_size
                        elif pruned_dim == IN_CHANNEL_DIM:
                            module.in_channels = pruned_size
                    elif isinstance(module, nn.Linear):
                        if pruned_dim == OUT_CHANNEL_DIM:
                            module.out_features = pruned_size
                        elif pruned_dim == IN_CHANNEL_DIM:
                            module.in_features = pruned_size
                    else:
                        pass
            pruned_dim += 1
    model_weight_names = model.state_dict().keys()
    pruned_weight_names = pruned_weights.keys()
    # check if module names match
    #assert set([prefix + _ for _ in model_weight_names]) == set(pruned_weight_names) , print (set([prefix + _ for _ in model_weight_names])  ,"\n\n\n", set(pruned_weight_names)) 
    
    #scratch:   ~    stem  
    #test :    stem  stem

    # inplace or return a new copy
    if not inplace:
        pruned_model = copy.deepcopy(model)
    else:
        pruned_model = model

    # update modules with mis-matched weight
    model_weights = pruned_model.state_dict()
    for model_weight_name in model_weight_names:
       # print(model_weight_name,":(",model_weights[model_weight_name].shape ,pruned_weights[prefix + model_weight_name].shape,")")
        if model_weights[model_weight_name].shape != pruned_weights[prefix + model_weight_name].shape:
            # print("   ",model_weight_name," ",prefix + model_weight_name)
            *container_names, module_name, param_name = model_weight_name.split('.')
            container = model
            for container_name in container_names:
                container = container._modules[container_name]
            module = container._modules[module_name]
            _dirty_fix(module, param_name, pruned_weights[prefix + model_weight_name].shape)

    # print(pruned_model.state_dict()['conf.3.weight'].shape ,pruned_weights['conf.3.weight'].shape)
    if load_pruned_weights:
        pruned_model.load_state_dict({k: v for k, v in pruned_weights.items()})
    return pruned_model


    

def rebuild(net,new_state_dict,load_pruned_weights=True):
    depth_wise_flag = []
    for k,v in net.state_dict().items():
        if 'weight' in k and len(v.shape)==4 and v.shape[0]==1:
            depth_wise_flag.append(  k[:-7]  )


    weights_m = new_state_dict
    from collections import OrderedDict
    new_state_dict = OrderedDict()
    for k, v in weights_m.items():
        head = k[:7]
        if head == 'module.':
            name = k[7:] # remove `module.`
        else:
            name = k
        new_state_dict[name] = v
    net =  load_pruned_model(net, new_state_dict,load_pruned_weights=load_pruned_weights) #load_pruned_weights = scratch flag
   
    

    for k,v in net.state_dict().items():
        if(torch.tensor(torch.tensor(v.shape).shape).item()) != 4:
            continue
        module_name = '.'.join(  k.split('.')[:-1]   )    
        param_name = k.split('.')[-1]
        container = get_module( net  , module_name)        
        if container.groups !=1:
            container.groups      = v.shape[OUT_CHANNEL_DIM]
            container.in_channels = v.shape[OUT_CHANNEL_DIM]
            container.out_channels  = v.shape[OUT_CHANNEL_DIM]

    return net




import inspect

__all__ = ["Registry"]


class Registry(object):
    def __init__(self, name):
        self._name = name
        self._module_dict = dict()

    def __repr__(self):
        format_str = self.__class__.__name__ + '(name={}, items={})'.format(
            self._name, list(self._module_dict.keys()))
        return format_str

    @property
    def name(self):
        return self._name

    @property
    def module_dict(self):
        return self._module_dict

    def get(self, key):
        return self._module_dict.get(key, None)

    def _register_module(self, module_class):
        module_name = module_class.__name__
        if module_name in self._module_dict:
            raise KeyError('{} is already registered in {}.'.format(
                module_name, self.name))
        self._module_dict[module_name] = module_class

    def register(self, cls):
        self._register_module(cls)
        return cls


from ...utils.flops_counter import get_model_complexity_info
import torch
import torch.nn as nn
from ..utils import get_pruning_mask, pruning_model_by_mask


class Uniform(object):
    def __init__(self, opt):
        #assert 输入参数是否有问题
        #
        self.model = opt.model
        self.input_shape = opt.input_shape
        import copy
        self.flops , self.parameter = get_model_complexity_info ( copy.deepcopy(self.model) , ( opt.input_shape ) , print_per_layer_stat=False )
        self.criterion = opt.prune_criterion
        self.save_flops = opt.prune_save_flops
        self.norm_layer_names = opt.norm_layer_names
        self.prec_layers = opt.prec_layers
        self.succ_layers = opt.succ_layers
        #self.source = {}
        self.weight = opt.model.state_dict()
        self.result = {}#run完后结果存在这里
        self.channel_numbers = {}
        self.opt = opt
        #for node_list in norm_layer_names:  ##获取
        #    weight  = []
        #    for i in node_list：
        #        weight.append(self.weights[i])
        #    self.source[name] = self.criterion(weight[name]))
        for node_list in self.norm_layer_names:
            self.channel_numbers[node_list[0]] = len(self.weight[node_list[0]+'.weight'])

    def run(self):
        """Get the pruning mask by the chennel_number_dict and importance criterion.
            
            Args:
                model: the torch net to be pruned.
                channel_number_dict(dict): the chennel_number_dict get by the fun{get_channel_numbers}
                criterion(str):  the importance criterion.
                        ['L1','L2','FPGM']
            Returns:
                pruning_mask(dict): the pruning mask,Like { 'layer1.bn':torch.Tensor([True,False,False,True,...]) , ....}
        """
        def equal_(a,b):
            return abs(a-b) <1e-5
        #读取outdir结果 uniform不用
        #resume uniform不用
        #二分选择合适的flops
        ll = 0.01
        rr = 0.99
        exit_flops = 0
        exit_flag  = 0
        while True:
            #import pdb;pdb.set_trace()
            mid = (ll+rr)/2
            #准备mask
            channel_savenumber_dict = {}
            for name  in self.norm_layer_names:          
                channel_savenumber_dict[ name[0] ] = int(mid * self.channel_numbers[name[0]] )
            
            pruning_mask = get_pruning_mask(self.opt , channel_savenumber_dict  )
           
            #prune net
            import copy
            net = copy.deepcopy(self.model)
            net = pruning_model_by_mask(net, pruning_mask, self.norm_layer_names, self.prec_layers, self.succ_layers)
            flops_pruned , _  =  get_model_complexity_info(copy.deepcopy(net), ( self.input_shape ) , print_per_layer_stat=False )
            #test pruned_net flops
            flops_ = flops_pruned  / self.flops * 100
            print('flops:',flops_)
            assert not equal_(flops_ , 100.0) and not equal_(flops_, 100.0) , 'flops_ error, exit binary search cycle'
            if abs(flops_ - self.save_flops * 100) < 1:
                self.result = channel_savenumber_dict
                break
            elif flops_ < self.save_flops * 100  :
                del net 
                ll = mid
            else:
                del net 
                rr = mid
            print("ll:",ll, ", rr:",rr,  ", mid:",mid, ", flops:", flops_)
            if exit_flops != flops_:
                exit_flops = flops_
                exit_flag = 0
            else:
                exit_flag +=1
                if exit_flag ==2:
                    print('warning !!!  Pruning Hyperparameters are not suitable.')
                    exit()




def get_module(model,name):
    assert type(name) is str
    
    if name.endswith('.bias'):
        name = name[:-5]
    if name.endswith('.weight'):
        name = name[:-7]
    
    container_names = name.split('.')
    if hasattr( model, 'module'):
        container = model.module
    else:
        container = model
    for container_name in container_names:
        container = container._modules[container_name]
    return container
 

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        pass
 
    try:
        import unicodedata
        unicodedata.numeric(s)
        return True
    except (TypeError, ValueError):
        pass
 
    return False


def is_transposed(model,name):
    assert type(name) is str
    if name.endswith('.bias'):
        name = name[:-5]
    if name.endswith('.weight'):
        name = name[:-7]
       
    container_names = name.split('.')
    if hasattr( model, 'module'):
        container = model.module
    else:
        container = model
    for container_name in container_names:
        container = container._modules[container_name]
    
    if hasattr( container, 'transposed'  ):
        return container.transposed
    else:
        print("?",name)
        return False;



import torch 
from  ..prune.importance  import   l2_norm

def display_layer(model, norm_layer_names,   sparsity_type = 'BN'):

    #loss_s_1=0
    #loss_s_2=0
    #loss_s_3=0
    #loss_s_4=0
    #loss_s_5=0
    #loss_s_6=0
    #loss_s_7=0
    #loss_s_8=0
    #loss_s_9=0
    #loss_s_10=0
    #loss_s_11=0
    #loss_s_12=0
    #loss_s_13=0
    loss_s = [0]*13
    sum_channel =0
    flag =0

    for norm_layer_name in norm_layer_names:
        *container_names, module_name = norm_layer_name.split('.')
        if hasattr(model,'module'):
            container = model.module
        else:
            container = model
        for container_name in container_names:
            container = container._modules[container_name]
        m = container._modules[module_name]
        if sparsity_type == 'lasso':#conv
            if m.transposed:
                source = l2_norm( m.weight.data , axis = 1)
            else :
                source = l2_norm( m.weight.data , axis = 0)
            for i in range(13):
                loss_s[i] += torch.sum( torch.tensor(source).lt(0.1**(i))).item()

        else :# BN
            if flag==0 and  m.weight.data.shape[0]<100 :
            #if norm_layer_name == 'layer1.7.bn1':
                print("\nbn_weight:\n",m.weight.data)
                if hasattr(m.weight, 'grad') :
                    if hasattr(m.weight.grad, 'data'):
                        print("\nbn_grad:\n",m.weight.grad.data)
                flag =1
            for i in range(13):
                loss_s[i] += torch.sum(m.weight.data.abs().lt(0.1**(i))).item()

        sum_channel += m.weight.data.shape[0]
     
    for i in range(13):
        loss_s[i] /=  sum_channel 

    print('________________________')
    print('|scores   |Proportion   |')
    for i in range(13):
        print('|<1e-%-2d   |%.8f   |'%(i,loss_s[i]))
    print('￣￣￣￣￣￣￣￣￣￣￣￣')
    

    for i in range(13):
        print("\r %f"% loss_s[i] ,end="")
        print(loss_s[i])   



def update_layer(model, norm_layer_names, factor=1e-4, scaler = False , mask_dict =None,regular_method='L1'):
    """sparsity the model by adding the sparsity Regularization grad
    Arguments:
        model (pytorch model): the model instance
        norm_layer_names (easypruner.): sparse layer names
        factor(float): the sparsity factor
        ngpu(boolen): needed in getting module by name
        mask_dict(dict): needed in mask sparsity
        regular_method(str): the regular method 
    step:
         1. get BNlayers by the norm_layer_names
         2. get the sparsity grad
         3. mutil the mask
    Returns:
        None
        the result return by
        opt.norm_layer_names 
        opt.prec_layers     
        opt.succ_layers

        a model instance with pruned structure (and weights if load_pruned_weights==True)
    """    
    if scaler :
        factor = factor * 65535
    
    for norm_layer_name in norm_layer_names:
        #get BNlayers by the norm_layer_names
        *container_names, module_name = norm_layer_name.split('.')
        if hasattr(model,'module') :
            container = model.module
        else:
            container = model
        
        for container_name in container_names:
            container = container._modules[container_name]
        m = container._modules[module_name]

        if mask_dict is not None:
            if type(mask_dict) is dict:
                if norm_layer_name not in mask_dict.keys():
                    continue
                mask = mask_dict[norm_layer_name]
            elif type(mask_dict) is float:
                len_ = m.weight.data.shape[0]
                mask = torch.tensor( int(mask_dict*len_)*[0] + (len_ - int( mask_dict*len_))*[1]  ).bool().cuda()


        #get the sparsity grad
        if regular_method=='L1':
            add_grad = factor * torch.sign(m.weight.data) # L1
        elif regular_method=='L2':
            add_grad = factor * m.weight.data # L2
        elif regular_method=='group_lasso':

            assert isinstance(m, (nn.Conv2d, nn.Conv3d, nn.ConvTranspose2d))
            if m.transposed:
                add_grad = m.weight.data * ((m.weight.data ** 2).sum(dim=(1, 2, 3), keepdim=True) ** (-0.5))  #  
            else:
                add_grad = m.weight.data * ((m.weight.data ** 2).sum(dim=(0, 2, 3), keepdim=True) ** (-0.5))  #              
        
        if  mask is not None :
            if regular_method=='group_lasso':
                assert len(mask.shape) == 4
            #mask = ~mask # mask the saved index   
            #add_grad = add_grad.mul(mask.float())
            add_grad[ mask.cuda() ] = 0 

        m.weight.grad.data.add_( add_grad )



def update_layer_grad_decay(model, norm_layer_names, optimizer = None, lr = None, 
                    scaler = False, mask_dict = None, epoch=0, epoch_decay = 0, iters =0):
                    
    """sparsity the model by adding the sparsity Regularization grad
    Arguments:
        model (pytorch model): the model instance
        norm_layer_names (easypruner.): sparse layer names
        optimizer: get the lr
        lr: one element or [lr_weight , lr_bias] . #we can use optimizer replace lr
        scaler(bool): if 
        mask_dict(dict): needed in mask sparsity
        epoch(int): Current epoch
        epoch_decay(int): the epochs when weight decay to zero
        iters(int): iters of one epoch
    step:
         1. get BNlayers by the norm_layer_names
         2. get the sparsity grad
         3. mutil the mask
    Returns:
        None
        the result return by
        opt.norm_layer_names 
        opt.prec_layers     
        opt.succ_layers

        a model instance with pruned structure (and weights if load_pruned_weights==True)
    """ 
    assert optimizer is not None or lr is not None
    if lr is not None:
        assert isinstance(lr,list)
        assert len(lr)>0 
        if len(lr)==1:
            lr.append(lr[0])
  
    assert mask_dict is not None ,"mask_dict needed"
    assert iters > 0
    assert epoch_decay >0
    factor = iters *(    epoch_decay  - epoch   )
    if epoch >=  (  epoch_decay  -2   )  :
        factor = 2*iters
    grad_decay = epoch/(  epoch_decay -20    )
    if grad_decay >=1.0:
        grad_decay=1.0
    
    s = 65535 if scaler else 1
   

    for norm_layer_name in norm_layer_names:
        #get BNlayers by the norm_layer_names
        *container_names, module_name = norm_layer_name.split('.')
        if hasattr(model,'module') :
            container = model.module
        else:
            container = model
        
        for container_name in container_names:
            container = container._modules[container_name]
        m = container._modules[module_name]
        
        if type(mask_dict) is dict:
            if norm_layer_name not in mask_dict.keys():
                continue
            mask = mask_dict[norm_layer_name]
        elif type(mask_dict) is float:
            len_ = m.weight.data.shape[0]
            mask = torch.tensor( int(mask_dict*len_)*[0] + (len_ - int( mask_dict*len_))*[1]  ).bool().cuda()


        w_l1 = []
        b_l1= []
        if optimizer is not None:
            lr = []
            flag =False
            for param_groups in optimizer.param_groups :
                for  param in param_groups['params']:
                    if m.weight is param:
                        lr.append(param_groups['lr'])
                        flag = True
                        break
                if flag:
                    break
            flag = False
            for param_groups in optimizer.param_groups :
                for param in param_groups['params']:
                    if m.bias is param:
                        lr.append(param_groups['lr'])
                        flag = True
                        break
                if flag:
                    break 


        for i in   range(m.weight.data.shape[0]):
            w_l1.append(   s*m.weight.data[i].item() / (lr[0] *factor )    )
            b_l1.append(   s*m.bias.data[i].item() / (lr[1] * factor )      )

        w_l1 = torch.tensor(w_l1).cuda()
        b_l1 = torch.tensor(b_l1).cuda()

        add_grad_w = w_l1
        add_grad_b = b_l1

        add_grad_w[mask.cuda()]=0
        add_grad_b[mask.cuda()]=0

        add_grad_w[~mask.cuda()]-= m.weight.grad.data[  ~mask.cuda() ] * grad_decay
        add_grad_b[~mask.cuda()]-= m.bias.grad.data[  ~mask.cuda() ]  * grad_decay
        #add_grad[0: int(add_grad.shape[0]/2)]=0
        if lr[0] !=0:
            m.weight.grad.data.add_( add_grad_w  )
        if lr[1] !=0:
            m.bias.grad.data.add_( add_grad_b  )



import numpy
import torch
from ...utils.judge import is_transposed 
__all__ = [  'OT_threshold'   ]
MIN_SCALING_FACTOR = 1e-18
OT_DISCARD_PERCENT = 0.0001
OUT_CHANNEL_DIM = 0
IN_CHANNEL_DIM = 1
WEIGHT_POSTFIX = ".weight"
BIAS_POSTFIX = ".bias"
MIN_CHANNELS = 1
OT_DISCARD_PERCENT = 1e-4
def OT_threshold(x, percent=OT_DISCARD_PERCENT):
    x_np = numpy.array(x)
    x_np[x_np < MIN_SCALING_FACTOR] = MIN_SCALING_FACTOR
    x_sorted = numpy.sort(x_np)
    #print("\n max:",x_sorted[-1])
    #print("\n",torch.from_numpy(x_sorted))
    if x_sorted[-1] <1e-3 :# no bigger peak, all channels have been sparsed , prune all
        th = 1
    elif x_sorted[-1] < x_sorted[0]*20  and x_sorted[-1] > 1e-3 :   #no smaller peak, all the channels are important,none will be pruned
        th = 0
    else:
        x2 = x_sorted**2
        Z = x2.sum()
        energy_loss = 0
        for i in range(x2.size):
            energy_loss += x2[i]
            if energy_loss / Z > percent:
                break
        th = (x_sorted[i-1] + x_sorted[i]) / 2 if i > 0 else 0
    return th


def get_mask_threshold(model, prune_factor , criterion , pruned_names, threshold_method = 'Order' ):
    """Get the pruning mask by criterion and method of threshold selecting.
    
    Args:
        model: the torch net to be pruned.
        prune_factor(float): the factor needed by threshold_mold
        criterion(str):  the importance criterion.
                   ['L1','L2','FPGM']
        threshold_method(str): method of threshold selecting.
                   ['order' , 'ratio' ,'OT']
                    1. order: direct order the threshold
                    2. ratio: getting threshold by channel saved ratio and global ranking channel 
                    3. OT: Optimal threshold method

    Returns:
        pruning_mask(dict): the pruning mask ,Like { 'layer1.bn':torch.Tensor([True,False,False,True,...]) , ....}
    """


  
    weights  = model.state_dict()
    scores_dict = {}
    for names in pruned_names :
        if  is_transposed(model, names[0]):
            scores_dict[names[0]] = criterion(weights[names[0]+WEIGHT_POSTFIX] , IN_CHANNEL_DIM )
        else :
            scores_dict[names[0]] = criterion(weights[names[0]+WEIGHT_POSTFIX] , OUT_CHANNEL_DIM )

    if threshold_method == 'Ratio':
        saved_numbers_dict = ratio_threshold_prune(  scores_dict,  prune_factor , criterion  )
    elif threshold_method == 'Order':
        saved_numbers_dict = order_threshold_prune( scores_dict , prune_factor , criterion )
    elif threshold_method == 'OT':
        saved_numbers_dict = ot_threshold_prune( scores_dict , prune_factor , criterion  )
    return saved_numbers_dict


def  ratio_threshold_prune( scores_dict, prune_factor , criterion  ):
    saved_numbers_dict = {}
    scores_list = []
    for name in scores_dict:
        scores_list += scores_dict[name]
    save_number = int( prune_factor * len( scores_list ))
    scores_list.sort(reverse=True)
    threshold =  scores_list[save_number]

    for name in scores_dict:
        scores = scores_dict[name]
        saved_numbers_dict[name] = (torch.tensor(scores) >threshold).sum()

    return saved_numbers_dict


def  order_threshold_prune( scores_dict, prune_factor , criterion  ):
    threshold = prune_factor
    saved_numbers_dict = {}
    for name in scores_dict:
        scores = scores_dict[name]
        saved_numbers_dict[name] = (torch.tensor(scores) >threshold).sum()
 
    return saved_numbers_dict  

def  ot_threshold_prune( scores_dict, prune_factor , criterion ):
    saved_numbers_dict = {}

    for name in scores_dict:
        scores = scores_dict[name]
        threshold = OT_threshold(scores.copy() , prune_factor)
        saved_numbers_dict[name] = (torch.tensor(scores) >threshold).sum()

    return saved_numbers_dict







import torch
import torch.nn as nn
import numpy as np
import copy
import functools


def get_model_flops(model, input_res, print_per_layer_stat=True, input_constructor=None):
    assert type(input_res) is tuple, 'Please provide the size of the input image.'
    assert len(input_res) >= 3, 'Input image should have 3 dimensions.'
    flops_model = add_flops_counting_methods(model)
    flops_model.eval().start_flops_count()
    if input_constructor:
        input = input_constructor(input_res)
        _ = flops_model(**input)
    else:
        device = list(flops_model.parameters())[-1].device
        batch = torch.FloatTensor(1, *input_res).to(device)
        if type(model.forward) is functools.partial:
            _ = flops_model([batch])
        else:
            _ = flops_model(batch)

    if print_per_layer_stat:
        print_model_with_flops(flops_model)
    flops_count = flops_model.compute_average_flops_cost()
    flops_model.stop_flops_count()

    return flops_count


def get_model_parameters(model):
    params_num = sum(p.numel() for p in model.parameters() if p.requires_grad)
    return params_num


def get_model_complexity_info(model0, input_res, print_per_layer_stat=True, as_strings=False, input_constructor=None):
    model = copy.deepcopy(model0)
    flops_count = get_model_flops(model, input_res, print_per_layer_stat, input_constructor)
    params_count = get_model_parameters(model)

    if as_strings:
        return flops_to_string(flops_count), params_to_string(params_count)

    return flops_count, params_count


def flops_to_string(flops, units='GMac', precision=2):
    if units is None:
        if flops // 10**9 > 0:
            return str(round(flops / 10.**9, precision)) + ' GMac'
        elif flops // 10**6 > 0:
            return str(round(flops / 10.**6, precision)) + ' MMac'
        elif flops // 10**3 > 0:
            return str(round(flops / 10.**3, precision)) + ' KMac'
        else:
            return str(flops) + ' Mac'
    else:
        if units == 'GMac':
            return str(round(flops / 10.**9, precision)) + ' ' + units
        elif units == 'MMac':
            return str(round(flops / 10.**6, precision)) + ' ' + units
        elif units == 'KMac':
            return str(round(flops / 10.**3, precision)) + ' ' + units
        else:
            return str(flops) + ' Mac'


def params_to_string(params_num):
    if params_num // 10 ** 6 > 0:
        return str(round(params_num / 10 ** 6, 2)) + ' M'
    elif params_num // 10 ** 3:
        return str(round(params_num / 10 ** 3, 2)) + ' k'
    else:
        return str(params_num)


def print_model_with_flops(model, units='GMac', precision=3):
    total_flops = model.compute_average_flops_cost()

    def accumulate_flops(self):
        if is_supported_instance(self):
            return self.__flops__ / model.__batch_counter__
        else:
            sum = 0
            for m in self.children():
                sum += m.accumulate_flops()
            return sum

    def flops_repr(self):
        accumulated_flops_cost = self.accumulate_flops()
        return ', '.join([flops_to_string(accumulated_flops_cost, units=units, precision=precision),
                          '{:.3%} MACs'.format(accumulated_flops_cost / total_flops),
                          self.original_extra_repr()])

    def add_extra_repr(m):
        m.accumulate_flops = accumulate_flops.__get__(m)
        flops_extra_repr = flops_repr.__get__(m)
        if m.extra_repr != flops_extra_repr:
            m.original_extra_repr = m.extra_repr
            m.extra_repr = flops_extra_repr
            assert m.extra_repr != m.original_extra_repr

    def del_extra_repr(m):
        if hasattr(m, 'original_extra_repr'):
            m.extra_repr = m.original_extra_repr
            del m.original_extra_repr
        if hasattr(m, 'accumulate_flops'):
            del m.accumulate_flops

    model.apply(add_extra_repr)
    print(model)
    model.apply(del_extra_repr)


def add_flops_counting_methods(net_main_module):
    # adding additional methods to the existing module object,
    # this is done this way so that each function has access to self object
    # embed()


    net_main_module.start_flops_count = start_flops_count.__get__(net_main_module)
    net_main_module.stop_flops_count = stop_flops_count.__get__(net_main_module)
    net_main_module.reset_flops_count = reset_flops_count.__get__(net_main_module)
    net_main_module.compute_average_flops_cost = compute_average_flops_cost.__get__(net_main_module)

    net_main_module.reset_flops_count()
    #import pdb;pdb.set_trace()
    # Adding variables necessary for masked flops computation
    net_main_module.apply(add_flops_mask_variable_or_reset)

    return net_main_module


def compute_average_flops_cost(self):
    """
    A method that will be available after add_flops_counting_methods() is called
    on a desired net object.

    Returns current mean flops consumption per image.

    """

    batches_count = self.__batch_counter__
    flops_sum = 0
    for module in self.modules():
        if is_supported_instance(module):
            flops_sum += module.__flops__

    return flops_sum / batches_count


def start_flops_count(self):
    """
    A method that will be available after add_flops_counting_methods() is called
    on a desired net object.

    Activates the computation of mean flops consumption per image.
    Call it before you run the network.

    """
    add_batch_counter_hook_function(self)
    self.apply(add_flops_counter_hook_function)



def stop_flops_count(self):
    """
    A method that will be available after add_flops_counting_methods() is called
    on a desired net object.

    Stops computing the mean flops consumption per image.
    Call whenever you want to pause the computation.

    """
    remove_batch_counter_hook_function(self)
    self.apply(remove_flops_counter_hook_function)


def reset_flops_count(self):
    """
    A method that will be available after add_flops_counting_methods() is called
    on a desired net object.

    Resets statistics computed so far.

    """
    add_batch_counter_variables_or_reset(self)
    self.apply(add_flops_counter_variable_or_reset)


def add_flops_mask(module, mask):
    def add_flops_mask_func(module):
        if isinstance(module, nn.Conv2d):
            module.__mask__ = mask
    module.apply(add_flops_mask_func)


def remove_flops_mask(module):
    module.apply(add_flops_mask_variable_or_reset)


# ---- Internal functions
def is_supported_instance(module):

    if isinstance(module, (nn.Conv2d, nn.Conv3d, nn.ConvTranspose2d,
                           nn.Linear,
                           nn.BatchNorm2d, nn.BatchNorm3d,
                           nn.ReLU, nn.PReLU, nn.ELU, nn.LeakyReLU, nn.ReLU6,
                           )):
        return True

    # if isinstance(module, (nn.Conv2d, nn.Conv3d, nn.Linear, common.DConv2d)):
    #     return True

    # if isinstance(module, (nn.Conv2d, nn.Conv3d, nn.ReLU, nn.PReLU, nn.ELU,
    #                        nn.LeakyReLU, nn.ReLU6, nn.Linear,
    #                        nn.MaxPool2d, nn.AvgPool2d, nn.BatchNorm2d,
    #                        nn.Upsample, nn.AdaptiveMaxPool2d, nn.AdaptiveAvgPool2d,
    #                        nn.ConvTranspose2d, nn.BatchNorm3d,
    #                        nn.MaxPool3d, nn.AvgPool3d, nn.AdaptiveMaxPool3d, nn.AdaptiveAvgPool3d)):
    #     return True

    return False


def empty_flops_counter_hook(module, input, output):
    module.__flops__ += 0


def upsample_flops_counter_hook(module, input, output):
    output_size = output[0]
    batch_size = output_size.shape[0]
    output_elements_count = batch_size
    for val in output_size.shape[1:]:
        output_elements_count *= val
    module.__flops__ += int(output_elements_count)


def relu_flops_counter_hook(module, input, output):
    active_elements_count = output.numel()
    module.__flops__ += int(active_elements_count)


def linear_flops_counter_hook(module, input, output):
    input = input[0]
    if len(input.shape) == 1:
        batch_size = 1
        module.__flops__ += int(batch_size * input.shape[0] * output.shape[0])
    else:
        batch_size = input.shape[0]
        module.__flops__ += int(batch_size * input.shape[1] * output.shape[1])


def pool_flops_counter_hook(module, input, output):
    input = input[0]
    module.__flops__ += int(np.prod(input.shape))


def bn_flops_counter_hook(module, input, output):
    module.affine
    input = input[0]
    # TODO: need to check here
    # batch_flops = np.prod(input.shape)
    # if module.affine:
    #     batch_flops *= 2
    # module.__flops__ += int(batch_flops)
    batch = input.shape[0]
    output_dims = input.shape[2:]
    channels = module.num_features
    batch_flops = batch * channels * np.prod(output_dims)
    if module.affine:
        batch_flops *= 2
    module.__flops__ += int(batch_flops)


def deconv_flops_counter_hook(conv_module, input, output):
    # Can have multiple inputs, getting the first one
    input = input[0]

    batch_size = input.shape[0]
    input_height, input_width = input.shape[2:]

    kernel_height, kernel_width = conv_module.kernel_size
    in_channels = conv_module.in_channels
    out_channels = conv_module.out_channels
    groups = conv_module.groups

    filters_per_channel = out_channels // groups
    conv_per_position_flops = kernel_height * kernel_width * in_channels * filters_per_channel

    active_elements_count = batch_size * input_height * input_width
    overall_conv_flops = conv_per_position_flops * active_elements_count
    bias_flops = 0
    if conv_module.bias is not None:
        output_height, output_width = output.shape[2:]
        bias_flops = out_channels * batch_size * output_height * output_height
    overall_flops = overall_conv_flops + bias_flops

    conv_module.__flops__ += int(overall_flops)


def dconv_flops_counter_hook(dconv_module, input, output):
    input = input[0]

    batch_size = input.shape[0]
    output_dims = list(output.shape[2:])

    m_channels, in_channels, kernel_dim1, _, = dconv_module.weight.shape
    out_channels, _, kernel_dim2, _, = dconv_module.projection.shape
    # groups = dconv_module.groups

    # filters_per_channel = out_channels // groups
    conv_per_position_flops1 = kernel_dim1 ** 2 * in_channels * m_channels
    conv_per_position_flops2 = kernel_dim2 ** 2 * out_channels * m_channels
    active_elements_count = batch_size * np.prod(output_dims)

    overall_conv_flops = (conv_per_position_flops1 + conv_per_position_flops2) * active_elements_count
    overall_flops = overall_conv_flops

    dconv_module.__flops__ += int(overall_flops)
    # dconv_module.__output_dims__ = output_dims


def conv_flops_counter_hook(conv_module, input, output):
    # embed()
    # Can have multiple inputs, getting the first one
    input = input[0]

    batch_size = input.shape[0]
    output_dims = list(output.shape[2:])

    kernel_dims = list(conv_module.kernel_size)
    in_channels = conv_module.in_channels
    out_channels = conv_module.out_channels
    groups = conv_module.groups

    filters_per_channel = out_channels // groups
    conv_per_position_flops = np.prod(kernel_dims) * in_channels * filters_per_channel

    active_elements_count = batch_size * np.prod(output_dims)

    overall_conv_flops = conv_per_position_flops * active_elements_count

    overall_flops = overall_conv_flops

    conv_module.__flops__ += int(overall_flops)
    conv_module.__output_dims__ = output_dims


# def conv_flops_counter_hook(conv_module, input, output):
#     # Can have multiple inputs, getting the first one
#     input = input[0]
#
#     batch_size = input.shape[0]
#     output_dims = list(output.shape[2:])
#
#     kernel_dims = list(conv_module.kernel_size)
#     in_channels = conv_module.in_channels
#     out_channels = conv_module.out_channels
#     groups = conv_module.groups
#
#     filters_per_channel = out_channels // groups
#     conv_per_position_flops = np.prod(kernel_dims) * in_channels * filters_per_channel
#
#     active_elements_count = batch_size * np.prod(output_dims)
#
#     if conv_module.__mask__ is not None:
#         # (b, 1, h, w)
#         flops_mask = conv_module.__mask__.expand(batch_size, 1, output_height, output_width)
#         active_elements_count = flops_mask.sum()
#
#     overall_conv_flops = conv_per_position_flops * active_elements_count
#
#     bias_flops = 0
#
#     if conv_module.bias is not None:
#
#         bias_flops = out_channels * active_elements_count
#
#     overall_flops = overall_conv_flops + bias_flops
#
#     conv_module.__flops__ += int(overall_flops)


def batch_counter_hook(module, input, output):
    batch_size = 1
    if len(input) > 0:
        # Can have multiple inputs, getting the first one
        input = input[0]
        batch_size = len(input)
    else:
        pass
        print('Warning! No positional inputs found for a module, assuming batch size is 1.')
    module.__batch_counter__ += batch_size


def add_batch_counter_variables_or_reset(module):

    module.__batch_counter__ = 0


def add_batch_counter_hook_function(module):
    if hasattr(module, '__batch_counter_handle__'):
        return

    handle = module.register_forward_hook(batch_counter_hook)
    module.__batch_counter_handle__ = handle


def remove_batch_counter_hook_function(module):
    if hasattr(module, '__batch_counter_handle__'):
        module.__batch_counter_handle__.remove()
        del module.__batch_counter_handle__


def add_flops_counter_variable_or_reset(module):
    if is_supported_instance(module):
        module.__flops__ = 0


def add_flops_counter_hook_function(module):
    if is_supported_instance(module):
        if hasattr(module, '__flops_handle__'):
            return

        if isinstance(module, (nn.Conv2d, nn.Conv3d)):
            handle = module.register_forward_hook(conv_flops_counter_hook)
        elif isinstance(module, nn.ConvTranspose2d):
            handle = module.register_forward_hook(deconv_flops_counter_hook)
        elif isinstance(module, nn.Linear):
            handle = module.register_forward_hook(linear_flops_counter_hook)
        elif isinstance(module, (nn.BatchNorm2d, nn.BatchNorm3d)):
            handle = module.register_forward_hook(bn_flops_counter_hook)
        elif isinstance(module, (nn.ReLU, nn.PReLU, nn.ELU, nn.LeakyReLU, nn.ReLU6)):
            handle = module.register_forward_hook(relu_flops_counter_hook)
        elif isinstance(module, (nn.AvgPool2d, nn.MaxPool2d, nn.AdaptiveMaxPool2d,
                                 nn.AdaptiveAvgPool2d, nn.MaxPool3d, nn.AvgPool3d,
                                 nn.AdaptiveMaxPool3d, nn.AdaptiveAvgPool3d)):
            handle = module.register_forward_hook(pool_flops_counter_hook)
        elif isinstance(module, nn.Upsample):
            handle = module.register_forward_hook(upsample_flops_counter_hook)
        else:
            handle = module.register_forward_hook(empty_flops_counter_hook)
        module.__flops_handle__ = handle


def remove_flops_counter_hook_function(module):
    if is_supported_instance(module):
        if hasattr(module, '__flops_handle__'):
            module.__flops_handle__.remove()
            del module.__flops_handle__
# --- Masked flops counting


# Also being run in the initialization
def add_flops_mask_variable_or_reset(module):
    if is_supported_instance(module):
        module.__mask__ = None



#utils.py
#arch的剪枝
import copy
from functools import partial
import torch
import torch.nn as nn
from .graph_parser import get_pruning_layers
import numpy

from sklearn.cluster import KMeans
import matplotlib.pyplot as plt

MIN_SCALING_FACTOR = 1e-18
OT_DISCARD_PERCENT = 0.0001
OUT_CHANNEL_DIM = 0
IN_CHANNEL_DIM = 1
WEIGHT_POSTFIX = ".weight"
BIAS_POSTFIX = ".bias"
MIN_CHANNELS = 1




def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        pass
 
    try:
        import unicodedata
        unicodedata.numeric(s)
        return True
    except (TypeError, ValueError):
        pass
 
    return False

def plt_weight_OT(model_weight_dict):
    #import pdb;pdb.set_trace()
    #print(model_weight_dict)
    for k, (weight , prune_th) in model_weight_dict.items():
        #import pdb;pdb.set_trace()
        prune_th = css_thresholding(weight) 
        weight  = numpy.array(weight)
        weight[ weight  < MIN_SCALING_FACTOR] = MIN_SCALING_FACTOR
        weight  = numpy.sort(weight)
        
    pass



def is_transposed(model,data_name):

    *container_names, module_name = data_name.split('.')
    container = model
    for container_name in container_names:
        container = container._modules[container_name]
    return container.transposed

 
 


def css_thresholding(x, percent=OT_DISCARD_PERCENT, method = 'OT'):
    x_np = numpy.array(x)
    x_np[x_np < MIN_SCALING_FACTOR] = MIN_SCALING_FACTOR
    x_sorted = numpy.sort(x_np)
    #print("\n max:",x_sorted[-1])
    #print("\n",torch.from_numpy(x_sorted))
    if method == 'th':
        th = percent
        #print("\nth-"+str(percent)+"  th:",th,"\n")
    elif method == 'OT':
        if x_sorted[-1] <1e-3 :# no bigger peak, all channels have been sparsed , prune all
            th = 1
        #elif x_sorted[-1] < x_sorted[0]*20  and x_sorted[-1] > 1e-3 :   #no smaller peak, all the channels are important,none will be pruned
        #    th = 0
        else:
            x2 = x_sorted**2
            Z = x2.sum()
            energy_loss = 0
            for i in range(x2.size):
                energy_loss += x2[i]
                if energy_loss / Z > percent:
                    break
            th = (x_sorted[i-1] + x_sorted[i]) / 2 if i > 0 else 0
        print("\nOT-"+str(percent)+"  th:",th,"\n") 
    return th


def group_weight_names(weight_names):
    grouped_names = {}
    for weight_name in weight_names:
        group_name = '.'.join(weight_name.split('.')[:-1])
        if group_name not in grouped_names:
            grouped_names[group_name] = [weight_name, ]
        else:
            grouped_names[group_name].append(weight_name)
    return grouped_names

######OT method

def liu2017(model,weights, percent, prec_layers, succ_layers, per_layer_normalization=False,get_mask=False,pruning_mask=None,method = 'OT' ):
    """default pruning method as described in:
            Zhuang Liu et.al., "Learning Efficient Convolutional Networks through Network Slimming", in ICCV 2017"

    Arguments:
        weights (OrderedDict): unpruned model weights
        percent (float): OT algorithm Super-parameter
        prec_layers (dict): mapping from BN names to preceding convs/linears
        succ_layers (dict): mapping from BN names to succeeding convs/linears
        per_layer_normalization (bool): if do normalization by layer

    Returns:
        pruned_weights (OrderedDict): pruned model weights
    """

    # find all scale weights in BN layers
    
    #get the prune_mask for further sparsity
    prune_mask_dict = None
    if get_mask: 
        from collections import OrderedDict
        prune_mask_dict = OrderedDict()
    
    #get weight and
    from collections import OrderedDict 
    plt_dict = OrderedDict()
    
    scale_weights = []
    deepwise_mask = {}#####用于储存deepwise卷积剪枝
    no_name_list = [] #用于存下第一回没有遍历到的deepwise
    norm_layer_names = list(set(succ_layers) & set(prec_layers))

    # print("norm_layer_names:\n",norm_layer_names)
    # print("\nsucc_layers(",type(succ_layers),"):\n",succ_layers)
    # print("\nprec_layers(",type(prec_layers),"):\n",prec_layers)

    # unpruned_norm_layer_names = list(set(succ_layers) ^ set(prec_layers))
    
    grouped_weight_names = group_weight_names(weights.keys())
    
    ###遍历每一layers
    for norm_layer_name in norm_layer_names:
        norm_weight_name = norm_layer_name + WEIGHT_POSTFIX
        print(norm_layer_name)
        scale_weight = weights[norm_weight_name].abs()
################优先级最高
        ####group卷积重新制作剪枝掩码
        # 假设deepwise卷积只有一个输入一个输出
        flag_no  = False
        flag_yes = False
        for prune_layer_name in prec_layers[norm_layer_name]: #遍历前继
            for weight_name in grouped_weight_names[prune_layer_name]:  # 遍历前继层成员
                if weight_name.endswith(WEIGHT_POSTFIX):  # endswith以“.weight”结尾
                    if (weights[weight_name].size(IN_CHANNEL_DIM) == 1):  # 判断该层是不是一个deepwise卷积
                        if (weight_name in deepwise_mask):
                            prune_indices, prune_mask = deepwise_mask[weight_name]
                            #修改in_channels,groups,out_channels
                            inum = int(prune_mask.sum().item())
                            aa = weight_name.split('.')
                            layer_name_ = ''
                            for ai in aa[:-1]:
                                if is_number(ai):
                                    layer_name_ +=( '[' + ai + ']')
                                else :
                                    layer_name_ +=('.'+ai)                         
   
                            exec('model'+layer_name_+'.groups='+'inum')
                            exec('model'+layer_name_+'.in_channels='+'inum')
                            flag_yes = True
                        else:
                            no_name_list.append(norm_layer_name)
                            flag_no = True
                    break
        if flag_no:
            #print("deepwise等待上一层的剪枝掩码:", norm_layer_name, weight_name )
            continue
        if not flag_yes :
#############优先级次高
            if per_layer_normalization:
                scale_weight = scale_weight / scale_weight.sum()
            scale_weight_list = [_.abs().item() for _ in list(scale_weight)]
            prune_th =   css_thresholding(scale_weight_list, percent=percent ,method = method)


            plt_dict[norm_layer_name] = [scale_weight_list , prune_th]  #
            
          
            #if(norm_layer_name == 'model_s32.model_s32.1.4'):
             #   prune_th = 0.010
            #print(norm_weight_name,"剪枝阈值:",prune_th)
            
            if pruning_mask is not None:
                prune_mask = pruning_mask[norm_layer_name]
            else: 
                prune_mask = scale_weight > prune_th#获取剪枝掩码，大于剪枝阈值保留
            
          
            if prune_mask.sum().item() == scale_weight.size(0): #要剪枝的通道数（输出）为零，则跳过
            
             
                #print("要剪枝的为0:", norm_layer_name)
                continue

            # in case not to prune the whole layer
            if prune_mask.sum() < MIN_CHANNELS:#   局部剪枝保护（这里是最小通道3，我们也可以改成按比例保留）,重新构建剪枝掩码
                
                scale_weight_list = [_.abs().item() for _ in list(scale_weight)]
                scale_weight_list.sort(reverse=True)
                prune_mask = scale_weight >= scale_weight_list[MIN_CHANNELS-1]
                if prune_mask.sum()  != MIN_CHANNELS :
                    import pdb;pdb.set_trace()
                    prune_mask = scale_weight > prune_th
                    prune_mask[0]=True
                   

            prune_indices = torch.nonzero(prune_mask).flatten()#nonzero得到数组中非零值的坐标


        # print(norm_layer_name)

        if get_mask:
            prune_mask_dict[norm_layer_name] = prune_mask
 
        #print(prune_indices , prune_mask)
        #tensor([ 0,  2,  3,  4,  5,  6,  7,  8, 10, 11, 12, 14, 15, 16, 17, 19, 20, 21,
        #22, 23, 25, 26, 27, 29, 30, 31, 32, 34, 35, 36, 37, 38, 39, 40, 42, 43,
        #44, 46, 47, 48, 49, 51, 52, 55, 57, 58, 60, 61, 62, 63]) tensor([ True, False,  True,  True,  True,  True,  True,  True,  True, False,
        # True,  True,  True, False,  True,  True,  True,  True, False,  True,
        # True,  True,  True,  True, False,  True,  True,  True, False,  True,
        # True,  True,  True, False,  True,  True,  True,  True,  True,  True,
        # True, False,  True,  True,  True, False,  True,  True,  True,  True,
        #False,  True,  True, False, False,  True, False,  True,  True, False,
        # True,  True,  True,  True])

        if  get_mask:
            pass 

        # 1. prune source normalization layer，先剪掉剪掉源层（BN）
        for weight_name in grouped_weight_names[norm_layer_name]:
            weights[weight_name] = weights[weight_name].masked_select(prune_mask)

        # 2. prune target succeeding conv/linear/... layers     剪掉源层（BN）后继层的（输入通道）
        for prune_layer_name in succ_layers[norm_layer_name]:   #遍历后继
            for weight_name in grouped_weight_names[prune_layer_name]: #遍历后继的各层(如bn层的 weights ，bias ，meaning ，var)
                if weight_name.endswith(WEIGHT_POSTFIX):#endswith以“.weight”结尾
                    #import pdb;pdb.set_trace()
                    transpose_flag = is_transposed(model,weight_name[:-6])
                    #print(weight_name)#,prune_indices)
                    
                    if (weights[weight_name].size(IN_CHANNEL_DIM) == 1):#deepwise卷积，剪枝掩码存入字典
                        # print("                        ",norm_layer_name,weight_name)
                        deepwise_mask[weight_name]=[prune_indices , prune_mask]
                        continue

                    if transpose_flag:#反卷积
                        weights[weight_name] = weights[weight_name].index_select(OUT_CHANNEL_DIM, prune_indices)#剪输入通道
                    else:#正常卷积
                        weights[weight_name] = weights[weight_name].index_select(IN_CHANNEL_DIM, prune_indices)#剪输入通道
        # 3. prune target preceding conv/linear/... layers    #剪掉源层（BN）前继层的（输入通道）也就是该modules的卷积
        for prune_layer_name in prec_layers[norm_layer_name]: #同上
            for weight_name in grouped_weight_names[prune_layer_name]:#同上
                if weight_name.endswith(WEIGHT_POSTFIX):#endswith以“.weight”结尾
                    transpose_flag = is_transposed(model,weight_name[:-6])
                    if transpose_flag:#反卷积
                        weights[weight_name] = weights[weight_name].index_select(IN_CHANNEL_DIM, prune_indices)#剪输出通道
                    else:#正常卷积
                        weights[weight_name] = weights[weight_name].index_select(OUT_CHANNEL_DIM, prune_indices)#剪输出通道
                elif weight_name.endswith(BIAS_POSTFIX):#endswith以“.bias”结尾
                    weights[weight_name] = weights[weight_name].index_select(0, prune_indices)#剪输出通道

    
    print(no_name_list)
    #处理之前没有遍历到的deepwise卷积
    for norm_layer_name in no_name_list:
        ####group卷积重新制作剪枝掩码
        # 假设deepwise卷积只有一个输入一个输出
        flag_no  = False
        for prune_layer_name in prec_layers[norm_layer_name]:  # 遍历前继
            for weight_name in grouped_weight_names[prune_layer_name]:  # 遍历前继层成员
                if weight_name.endswith(WEIGHT_POSTFIX):  # endswith以“.weight”结尾
                    if(weights[weight_name].size(IN_CHANNEL_DIM) == 1):#判断该层是不是一个deepwise卷积
                        if (weight_name in deepwise_mask):
                            prune_indices, prune_mask = deepwise_mask[weight_name]
                            #修改in_channels,groups,out_channels
                            inum = int(prune_mask.sum().item())
                            aa = weight_name.split('.')
                            layer_name_ = ''
                            for ai in aa[:-1]:
                                if is_number(ai):
                                    layer_name_ +=( '[' + ai + ']')
                                else :
                                    layer_name_ +=('.'+ai)
                            exec('model'+layer_name_+'.groups='+'inum')
                            exec('model'+layer_name_+'.in_channels='+'inum')
                            # exec("print('model.'+layer_name_+'.groups')")
                        else:
                            flag_no = True
        if(flag_no):
            print(norm_layer_name,":daaddddasdasdasdasdasdasdasda!!!!!")
            continue


        print(norm_layer_name)
        #print(prune_indices , prune_mask)
        # tensor([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17,
        #         18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31])
        # tensor([True, True, True, True, True, True, True, True, True, True, True, True,
        #         True, True, True, True, True, True, True, True, True, True, True, True,
        #         True, True, True, True, True, True, True, True])




        # 1. prune source normalization layer，先剪掉剪掉源层（BN）
        for weight_name in grouped_weight_names[norm_layer_name]:
            weights[weight_name] = weights[weight_name].masked_select(prune_mask)

        # 2. prune target succeeding conv/linear/... layers     剪掉源层（BN）后继层的（输入通道）
        for prune_layer_name in succ_layers[norm_layer_name]:   #遍历后继
            for weight_name in grouped_weight_names[prune_layer_name]: #遍历后继的各层(如bn层的 weights ，bias ，meaning ，var)
                if weight_name.endswith(WEIGHT_POSTFIX):#endswith以“.weight”结尾
                    # print(weight_name,prune_indices)
                    if (weights[weight_name].size(IN_CHANNEL_DIM) == 1):#deepwise卷积，剪枝掩码存入字典
                        deepwise_mask[weight_name]=[prune_indices , prune_mask]
                        continue
                    transpose_flag = is_transposed(model,weight_name[:-6])
                    if transpose_flag:#反卷积
                        weights[weight_name] = weights[weight_name].index_select(OUT_CHANNEL_DIM, prune_indices)#剪输入通道
                    else:
                        weights[weight_name] = weights[weight_name].index_select(IN_CHANNEL_DIM, prune_indices)#剪输入通道
                       
        # 3. prune target preceding conv/linear/... layers    #剪掉源层（BN）前继层的（输入通道）也就是该modules的卷积
        for prune_layer_name in prec_layers[norm_layer_name]: #同上
            for weight_name in grouped_weight_names[prune_layer_name]:#同上
                if weight_name.endswith(WEIGHT_POSTFIX):#endswith以“.weight”结尾
                    transpose_flag = is_transposed(model,weight_name[:-6])
                    if transpose_flag:#反卷积
                        weights[weight_name] = weights[weight_name].index_select(IN_CHANNEL_DIM, prune_indices)#剪输出通道
                    else:
                        weights[weight_name] = weights[weight_name].index_select(OUT_CHANNEL_DIM, prune_indices)#剪输出通道
                elif weight_name.endswith(BIAS_POSTFIX):#endswith以“.bias”结尾
                    weights[weight_name] = weights[weight_name].index_select(0, prune_indices)#剪输出通道

     
    #plt_weight_OT( plt_dict  )

    return prune_mask_dict , weights


liu2017_normalized_by_layer = partial(liu2017, per_layer_normalization=True)


def _dirty_fix(module, param_name, pruned_shape):
    module_param = getattr(module, param_name)

    # identify the dimension to prune
    pruned_dim = 0
    for original_size, pruned_size in zip(module_param.shape, pruned_shape):
        if original_size != pruned_size:
            keep_indices = torch.LongTensor(range(pruned_size)).to(module_param.data.device)
            module_param.data = module_param.data.index_select(pruned_dim, keep_indices)

            # modify number of features/channels
            if param_name == "weight":
                if isinstance(module, nn.modules.batchnorm._BatchNorm) or \
                        isinstance(module, nn.modules.instancenorm._InstanceNorm) or \
                        isinstance(module, nn.GroupNorm):
                    module.num_features = pruned_size
                elif isinstance(module, nn.modules.conv._ConvNd):
                    if pruned_dim == OUT_CHANNEL_DIM:
                        module.out_channels = pruned_size
                    elif pruned_dim == IN_CHANNEL_DIM:
                        module.in_channels = pruned_size
                elif isinstance(module, nn.Linear):
                    if pruned_dim == OUT_CHANNEL_DIM:
                        module.out_features = pruned_size
                    elif pruned_dim == IN_CHANNEL_DIM:
                        module.in_features = pruned_size
                else:
                    pass
        pruned_dim += 1


def load_pruned_model(model, pruned_weights, prefix='', load_pruned_weights=True, inplace=True):
    """load pruned weights to a unpruned model instance

    Arguments:
        model (pytorch model): the model instance
        pruned_weights (OrderedDict): pruned weights
        prefix (string optional): prefix (if has) of pruned weights
        load_pruned_weights (bool optional): load pruned weights to model according to the ICLR 2019 paper:
            "Rethinking the Value of Network Pruning", without finetuning, the model may achieve comparable or even
            better results
        inplace (bool, optional): if return a copy of the model

    Returns:
        a model instance with pruned structure (and weights if load_pruned_weights==True)
    """
    model_weight_names = model.state_dict().keys()
    pruned_weight_names = pruned_weights.keys()

    # check if module names match
    assert set([prefix + _ for _ in model_weight_names]) == set(pruned_weight_names) , print (set([prefix + _ for _ in model_weight_names])  ,"\n\n\n", set(pruned_weight_names)) 
    
    #scratch:   ~    stem  
    #test :    stem  stem

    # inplace or return a new copy
    if not inplace:
        pruned_model = copy.deepcopy(model)
    else:
        pruned_model = model

    # update modules with mis-matched weight
    model_weights = pruned_model.state_dict()
    for model_weight_name in model_weight_names:
       # print(model_weight_name,":(",model_weights[model_weight_name].shape ,pruned_weights[prefix + model_weight_name].shape,")")
        if model_weights[model_weight_name].shape != pruned_weights[prefix + model_weight_name].shape:
            # print("   ",model_weight_name," ",prefix + model_weight_name)
            *container_names, module_name, param_name = model_weight_name.split('.')
            container = model
            for container_name in container_names:
                container = container._modules[container_name]
            module = container._modules[module_name]
            _dirty_fix(module, param_name, pruned_weights[prefix + model_weight_name].shape)

    # print(pruned_model.state_dict()['conf.3.weight'].shape ,pruned_weights['conf.3.weight'].shape)


    if load_pruned_weights:
        pruned_model.load_state_dict({k: v for k, v in pruned_weights.items()})
    return pruned_model


def prune(model, input_shape,  prune_ratio=OT_DISCARD_PERCENT, prune_method=liu2017, get_mask=False, pruning_mask = None, min_channels = 1 ,Threshold_method = 'th' ):
    """prune a model

    Arguments:
        model (pytorch model): the model instance
        input_shape (tuple): shape of the input tensor
        prune_ratio (float): ratio of be pruned channels to total channels
        prune_method (method): algorithm to prune weights

    Returns:
        a model instance with pruned structure (and weights if load_pruned_weights==True)

    Pipeline:
        1. generate mapping from tensors connected to BNs by parsing torch script traced graph
        2. identify corresponding BN and conv/linear like:
            conv/linear --> ... --> BN --> ... --> conv/linear
                                     |
                                    ...
                                     | --> relu --> ... --> conv/linear
                                    ...
                                     | --> ... --> maxpool --> ... --> conv/linear
            , where ... represents per channel operations. all the floating nodes must be conv/linear
        3. prune the weights of BN and connected conv/linear
        4. load weights to a unpruned model with pruned weights
    """
    global MIN_CHANNELS
    MIN_CHANNELS = min_channels

    # convert to CPU for simplicity
    src_device = next(model.parameters()).device
    model = model.cpu()

    # parse & generate mappings to BN layers
    prec_layers, succ_layers = get_pruning_layers(model, input_shape)
    # print(prec_layers)
    # print("\n",succ_layers)
    # prune weights
    prune_mask_dict,pruned_weights = prune_method(model=model,weights=model.state_dict(),  percent=prune_ratio, prec_layers = prec_layers,succ_layers=succ_layers,get_mask=get_mask,pruning_mask = pruning_mask, method = Threshold_method)
    #liu2017(model,weights, percent, prec_layers, succ_layers, per_layer_normalization=False,get_mask=False):
    # prune model according to pruned weights
    if MIN_CHANNELS ==0:
        return prune_mask_dict , pruned_weights
    else :
        pruned_model = load_pruned_model(model, pruned_weights)
        if get_mask:
            return prune_mask_dict , pruned_model.to(src_device)
        else:
            return pruned_model.to(src_device)




import os
from  .prune_algorithm import prune_finetune ,sparsity_prune_finetune, prune_op
from  .prune.importance import l1_norm,   l2_norm, geometry_median
from .utils.flops_counter import get_model_complexity_info
from .utils.judge import get_module 
import torch 

from .utils.get_prune_layer import get_prune_layers

class config(object):
    def __init__(self,exp_name, algorithm):
        #通用部分
        self.model = None #模型
        self.args = None #训练测试的参数
        self.train = None #train()函数              #### train(net,args)     return  None
        self.sparsity = None #sparsity()函数        #### sparsity(net,args)  return  None
        self.eval = None #eval()函数                #### eval(net,args)      return  acc_top1
        self.dataloader_train = None 
        self.dataloader_eval  = None  
        self.input_shape = None    #input_shape   (3,256,256)
        self.add_layer_need = True
        self.norm_layer_names = [] #(list)norm_layer_names:[ [namea] , [nameb] , [namec,named,..],..]
        self.prec_layers = []      #(dict)prec_layers { [name_a]:[name1] ,....[name_c,name_d]:[name_b,name_b,name_b1]      }
        self.succ_layers = []      #(dict)succ_layers { [name_a]:[name11],....[name_c,name_d];[name_b,name_b2,name_b3] } 
        self.prune_stride_2 = True  # 是否剪枝 stride为2的卷积层
        self.exp_name =exp_name  
        self.algorithm = algorithm #prune_algorithm_name
        self.workdir ='./workdir/'+exp_name
        if not os.path.exists(self.workdir) :
            os.makedirs(self.workdir)
        #pruning state
        self.pruned = False
        self.prunedmodel_flops = None
        self.prunedmodel_params = None
        self.real_prune_save_flops =None  
        #剪枝  
        self.prune_save_flops  = None    #save_flops float
        self.prune_factor  = None    # 阈值剪枝方法通道剪枝率
        self.prune_criterion   = None    #criterion()   importance函数
        self.axis = 1
        self.prune_mask = None   #剪枝算法
        if self.algorithm in ['Ratio','DDPG' , 'Order','EagleEye', 'Uniform', 'seneity_analysis']:
            self.prune_method  = self.algorithm  #剪枝mask
        else:
            self.prune_method = None 
        self.use_conv_statistics  = False # 默认False使用BN作为 importance
        self.dense_task = None
        #EagleEye
        self.EagleEye_search_nums= 1000
        self.EagleEye_max_rate = 0.7
        self.EagleEye_min_rate = 0.0
        self.Adaptive_BN = None 
        #稀疏部分
        self.sparsity_factor = 0   #稀疏因子
        self.sparsity_method = None#稀疏方法
        self.sparsity_mask = None  #None 是全局稀疏，mask限定稀疏范围
        self.sparsity_layernames =  []
 
        #finetune部分
        self.fientune_methhod = None #finetune算法
        #onnx_file 
        self.onnx_file = None

    def init(self):
        
        #get_prune_layers
        if len(self.norm_layer_names)==0  or len(self.prec_layers)==0  or  len(self.succ_layers)==0  :
            get_prune_layers(self)
        #get_importance
        if self.prune_criterion in ['L1' , 'l1' ]:
            self.prune_criterion = l1_norm
        elif self.prune_criterion in ['L2' , 'l2' ]:
            self.prune_criterion = l2_norm
        elif self.prune_criterion in ['FPGM' ]:
            self.prune_criterion = geometry_median

        #get_flops,params
        import copy  
        self.flops , self.params =   get_model_complexity_info( self.model , ( self.input_shape ) , print_per_layer_stat=False )

        #pruned_state
        self.pruned = False

        #pruned_mask
        if self.prune_mask:
            self.prune_mask = torch.load(self.prune_mask)
       
        #merge depthwise
        def get_index(name , name_list ):
            for index,  names  in enumerate( name_list ):
                if name in names:
                    return index 
        prec_layers_ = []
        norm_layer_names_ = []
        succ_layers_ = []
        continue_index = []
        for index,  conv_succlist in enumerate(self.succ_layers ):
            if index in continue_index:
                continue
            prec_layer = self.prec_layers[index].copy()
            norm_layer_name = self.norm_layer_names[index].copy()
            succ_layer = self.succ_layers[index].copy()
            for conv_name in conv_succlist :
                container = get_module(self.model , conv_name)
                if not  hasattr(container, 'groups') :
                    print(  conv_name ,"continue"  )
                    continue
                assert  hasattr(container, 'groups')
                assert  hasattr(container, 'out_channels')
                if container.groups ==container.out_channels:
                    index_ = get_index( conv_name , self.prec_layers )
                    #import pdb;pdb.set_trace()
                    continue_index.append( index_ )
                    prec_layer += self.prec_layers[index_]
                    norm_layer_name += self.norm_layer_names[index_]
                    succ_layer +=  self.succ_layers[index_]
            prec_layers_.append( prec_layer )
            norm_layer_names_.append( norm_layer_name )
            succ_layers_.append( succ_layer  )
                 
        self.prec_layers = prec_layers_
        self.norm_layer_names = norm_layer_names_
        self.succ_layers = succ_layers_
        
        #stride =2
        prec_layers_ = []
        norm_layer_names_ = []
        succ_layers_ = []
        if not self.prune_stride_2 :
            for index,convlist in enumerate( self.prec_layers ):
                flag = False
                for conv_name in convlist:
                    container = get_module(self.model , conv_name)
                    assert  hasattr(container, 'stride')
                    if container.stride==(2,2):
                        flag = True
                        break
                if flag:
                    continue
                else:
                    prec_layers_.append( self.prec_layers[index] )
                    norm_layer_names_.append(self.norm_layer_names[index])
                    succ_layers_.append( self.succ_layers[index])
            
            self.prec_layers = prec_layers_ 
            self.norm_layer_names = norm_layer_names_ 
            self.succ_layers = succ_layers_ 
        
        # groups != out_chanel 
        prec_layers_ = []
        norm_layer_names_ = []
        succ_layers_ = []
        for index,  conv_preclist  in enumerate(self.prec_layers ):
            flag = True
            conv_succlist = self.succ_layers[index]
            for conv_name in conv_succlist + conv_preclist:
                container = get_module(self.model , conv_name)
                if not  hasattr(container, 'groups') :
                    print(  conv_name ,"continue"  )
                    continue
                assert  hasattr(container, 'groups')
                assert  hasattr(container, 'out_channels')
                if container.groups !=1 and container.groups !=container.out_channels:
                    flag = False
                    print("Warning!!!!  groups != out_channels not support!!!!")
                    break
            if flag:
                prec_layers_.append( self.prec_layers[index] )
                norm_layer_names_.append(self.norm_layer_names[index])
                succ_layers_.append( self.succ_layers[index])
        self.prec_layers = prec_layers_ 
        self.norm_layer_names = norm_layer_names_ 
        self.succ_layers = succ_layers_ 
   

        print("BN",self.norm_layer_names)
        print("prec_conv",self.prec_layers)
        print("succ_conv",self.succ_layers)

        #sparisty
        for i in self.norm_layer_names:
            self.sparsity_layernames += i






        

    def prune_algorithm(self):
        if self.algorithm in ['Uniform', 'EagleEye']:
            prune_finetune(self)
        elif self.algorithm in ['netslim', 'maskl1']:
            sparsity_prune_finetune(self)

    def  prune_algorithm_onlyprune(self):
        if self.algorithm in ['Uniform', 'Ratio',  'Order']:
            prune_op(self)



   



import logging
import numpy as np
import torch 
#from ..common import get_logger
#from ..core import Registry, GraphWrapper

__all__ = ["l1_norm", "l2_norm","geometry_median"]

#_logger = get_logger(__name__, level=logging.INFO)

#CRITERION = Registry('criterion')





#@CRITERION.register
def l1_norm(weights, axis = 0):
    """Compute l1-norm scores of parameter on given axis.
    This function return a list of parameters' l1-norm scores on given axis.

    Each element of list is a tuple with format (name, axis, score) in which 'name' is parameter's name
    and 'axis' is the axis reducing on and `score` is a np.array storing the l1-norm of strucure on `axis`.
    
    Args:
        weights(): A group of parameters. the wights of BN layers / conv layers / deconv layers
        axis(int): axis is the axis of calculating scores. BN layers
        L1 :   Σabs(all)
    Returns:
       list: A list of tuple l1-norm on given axis.
    """
    assert axis < len(weights.shape)
    scores = []
    
    for i in range(weights.shape[axis]):
        weight =  (weights.index_select(axis, torch.tensor([i])).abs()).sum()
        scores.append(weight.item())

    return scores






#@CRITERION.register
def l2_norm(weights, axis = 0):
    """Compute l1-norm scores of parameter on given axis.
    This function return a list of parameters' l2-norm scores on given axis.

    Each element of list is a tuple with format (name, axis, score) in which 'name' is parameter's name
    and 'axis' is the axis reducing on and `score` is a np.array storing the l2-norm of strucure on `axis`.
    
    Args:
        weights(): A group of parameters. the wights of BN layers / conv layers / deconv layers
        axis(int): axis is the axis of calculating scores. BN layers
        L2 :   sqrt（Σ all**2 ）
    Returns:
       list: A list of tuple l1-norm on given axis.
    """
    assert axis < len(weights.shape)

    scores = []
    
    for i in range(weights.shape[axis]):
        weight =  (weights.index_select(axis, torch.tensor([i])).pow(2)).sum()
        weight = weight.sqrt()
        scores.append(weight.item())

    return scores





#@CRITERION.register
def geometry_median(weights, axis = 0 ):
    assert axis < len(weights.shape)
    
    scores = []
    dim_che = list(range(len(weights.shape)))
    dim_che[axis] = 0
    dim_che[0] = axis

    def get_distance_sum(weights, index, axis):
        weights_ = weights.clone()
        selected_filter = weights_.index_select(axis, torch.tensor([i])).permute(dim_che).reshape(1,-1).repeat( weights_.shape[axis] ,1 )
        weights_ = weights_.permute(dim_che).reshape(weights_.shape[axis] ,-1 )
        x = weights_ - selected_filter
        x = (x*x).sum(1).sqrt()
        return x.sum()

    dist_sum_list = []
    for i in range(weights.shape[axis]):
        score = get_distance_sum(weights, i, axis)
        scores.append(score)

    return scores







from ...utils.flops_counter import get_model_complexity_info
import torch
import torch.nn as nn
from ..utils import get_mask_layers, pruning_model_by_mask
import os
import numpy as np
import collections
import copy
import os.path
class EagleEye(object):
    def __init__(self, opt):
        #assert 输入参数是否有问题
        self.model = opt.model
        self.input_shape = opt.input_shape
        self.flops , self.parameter = get_model_complexity_info ( self.model , ( opt.input_shape ) , print_per_layer_stat=False )
        self.flag = 'flops'
        self.criterion = opt.prune_criterion
        self.save_flops = opt.prune_save_flops
        self.norm_layer_names = opt.norm_layer_names
        self.prec_layers = opt.prec_layers
        self.succ_layers = opt.succ_layers
        self.max_rate = opt.EagleEye_max_rate     
        self.min_rate = opt.EagleEye_min_rate       
        self.eval = opt.eval
        self.Adaptive_BN = opt.Adaptive_BN
        self.eval_args = opt.args
        self.weight = opt.model.state_dict()
        ###### create the search txt
        self.search_result_path = os.path.join( opt.workdir, opt.exp_name+ '_'+str(self.save_flops)+'flops.txt')
        self.serach_file= None 
        self.search_nums = opt.EagleEye_search_nums
    
        ##run完后结果存在这里
        self.result = {}
    
        ##Adaptive_BN
        self.dataloader_train = opt.dataloader_train

        self.channel_numbers = {}
        for node_list in self.norm_layer_names:
            self.channel_numbers[node_list[0]] = len(self.weight[node_list[0]+'.weight'])

    def run(self):
        """Get the pruning mask by the chennel_number_dict and importance criterion.
            Args:
                model: the torch net to be pruned.
                channel_number_dict(dict): the chennel_number_dict get by the fun{get_channel_numbers}
                criterion(str):  the importance criterion.
                        ['L1','L2','FPGM']
            Returns:
                pruning_mask(dict): the pruning mask,Like { 'layer1.bn':torch.Tensor([True,False,False,True,...]) , ....}
        """
        ################################################################################################################################# process the file
        #write the  layername in order
        if not os.path.isfile( self.search_result_path):
            self.serach_file= open( self.search_result_path , "w")
            for names  in self.norm_layer_names:
                self.serach_file.write("{}".format(names[0]) )
                for i in names[1:]:
                    self.serach_file.write(",{}".format(i) )
                self.serach_file.write(" ")
            self.serach_file.write("\n")
            self.serach_file.close() 
        self.serach_file= open( self.search_result_path , "r")
        lines = self.serach_file.readlines()  # 读取所有行
        len_results = len(lines)
        channel_number_dict = collections.OrderedDict()
        names_ = [  [j  for j in i.split(',')]    for i in lines[0].split( )] 
        assert set( [ i[0]  for i in names_] ) ==set( [ i[0]  for i in self.norm_layer_names] ) , "the layernames in file dont match the self.norm_layer_names"
        self.norm_layer_names = names_ 
        self.serach_file.close()
        ############################################################################################################################## pruning process
        while(len_results < 1+self.search_nums ):
            #########################search pruning strategy 
            channel_config = np.random.rand(len(self.norm_layer_names))
            channel_config = channel_config * self.max_rate
            channel_config = channel_config + self.min_rate
            channel_config = np.ones(len(self.norm_layer_names)) - channel_config
            pruning_strategy =channel_config.tolist() 
            #############################pruning model and judge the flop
            channel_savenumber_dict = {}
            for name, ratio  in zip(self.norm_layer_names , pruning_strategy):
                channel_savenumber_dict[name[0]] = int(ratio *self.channel_numbers[ name[0] ])
            #import pdb;pdb.set_trace()
            #print(pruning_strategy)
            net = copy.deepcopy(self.model)
            pruning_mask = get_mask_layers(self.model, self.criterion, channel_savenumber_dict ,self.norm_layer_names )
            try:
                net = pruning_model_by_mask( net  , pruning_mask , self.norm_layer_names , self.prec_layers , self.succ_layers)
            except:
                del net
                continue
            flops_pruned , parameter_pruned  =  get_model_complexity_info(net, ( self.input_shape ) , print_per_layer_stat=False )
            if self.flag == 'flops':
                flops_ = flops_pruned  / self.flops * 100
                print('flops:',flops_)
                if abs(flops_ - self.save_flops * 100) > 1:
                    del net 
                    continue
            elif self.flag == 'params':
                param_ = parameter_pruned  / self.parameter * 100
                print('param:',param_)
                if abs(param_ - self.save_flops * 100) > 1:
                    del net 
                    continue
            #########################################Adaptive BN
            self.Adaptive_BN(net,  self.need_dict)
            #########################################Test and record the pruning results
            strategy_score = self.eval(net,  self.need_dict)
            self.serach_file= open( self.search_result_path , "a+")
            self.serach_file.write("{} {} {} ".format(strategy_score, flops_ , param_))
            for item in pruning_strategy:
                self.serach_file.write("{} ".format(str(item)))
            self.serach_file.write("\n")
            self.serach_file.close()
            print("Eval Score:{}".format(strategy_score))

            # the result of searching add one
            del net  
            len_results +=1
        
        ####################################################################################################### Complete the search and select the best results
        self.get_best()


    #def Adaptive_BN(self, net , batchs =100,dataloader_train):
    #     from torch.autograd import Variable
    #     net.train()
    #     with torch.no_grad():
    #         for batch_index, (images, labels) in enumerate(self.dataloader_train ):
    #                 images = Variable(images)
    #                 _ = net(images)
    #                 if batch_index > batchs :
    #                     break
         
    
    def get_best(self):
        self.serach_file= open( self.search_result_path , "r")
        lines = self.serach_file.readlines()[1:]
        best_acc = 0.000000000
        best_index = 0
        for index, strings  in enumerate(lines):
            acc =  float(strings.strip().split()[0])
            if best_acc < acc :
                best_index = index
                best_acc = acc
        print('pruning_strategy_id:',best_index,'\npruning_strategy:',lines[best_index])
        pruning_strategy = [ float(i) for i in lines[best_index].split() ]
        channel_savenumber_dict = {}
        for name, ratio  in zip(self.norm_layer_names , pruning_strategy[2:]):
                channel_savenumber_dict[name[0]] = int( ratio * self.channel_numbers[ name[0] ])
        self.result = channel_savenumber_dict
        self.serach_file.close()



























import torch
import torch.nn as nn
from ...utils.rebuild import rebuild
from ...utils.judge import is_transposed

__all__ = ["get_pruning_mask", "pruning_model_by_mask"]


def get_pruning_mask(opt ,channel_savenumber_dict ):
    #opt.model, opt.prune_criterion ,channel_number_dict ,  opt.norm_layer_names, opt.prec_layers ,  opt.use_conv_statistics
    model = opt.model 
    criterion = opt.prune_criterion 
    norm_layer_names = opt.norm_layer_names
    prec_layers = opt.prec_layers 
    use_conv_statistics = opt.use_conv_statistics
 
    if use_conv_statistics:
        channel_savenumber_dict_ = {}
        for norm_layer_name , prec_layer in zip( norm_layer_names , prec_layers ):
            channel_savenumber_dict_[ prec_layer[0]] = channel_savenumber_dict[  norm_layer_name[0]]
        pruning_mask_ = get_mask_layers_conv(model, criterion, channel_savenumber_dict_ , prec_layers  )
        pruning_mask = {}
        for norm_layer_name , prec_layer in zip( norm_layer_names , prec_layers ):
            pruning_mask[ norm_layer_name[0]] = pruning_mask_[ prec_layer[0]]
    else:
        pruning_mask = get_mask_layers(model, criterion, channel_savenumber_dict ,  norm_layer_names)
    return pruning_mask

    

def get_mask_layers_conv(model, criterion, channel_savenumber_dict , pruned_names ):
    pass
    """
    Binary search to find the appropriate pruning rate.
        self.result get the channel_savenumber_dict 
    """
    pruning_mask = {}
    weights  = model.state_dict()
    for lid, names in enumerate(pruned_names):
        name = names[0]
        save_number  =  channel_savenumber_dict[name]
        weight = weights[name + '.weight'] #residual connecation choose the first layer// stride=2 layer
        scores = []
        if  is_transposed( model , name ): 
            scores = criterion( weight , 1 )
        else:
            scores = criterion( weight , 0 )
        scores_ = scores.copy()
        scores_.sort(reverse=True)
        if save_number == len(scores):
            threshold = scores_[save_number-1]-1
        else:
            threshold = scores_[save_number]
        pruning_mask[name]  = (torch.tensor(scores) >threshold).bool()
        if pruning_mask[name].sum() ==0 :
            pruning_mask[name]  =( torch.tensor(scores) == scores[0]    ).bool()
    return pruning_mask
    

def get_mask_layers(model, criterion, channel_savenumber_dict , pruned_names  ):
    """Binary search to find the appropriate pruning rate.
        self.result get the channel_savenumber_dict 
    """

    pruning_mask = {}
    weights  = model.state_dict()
    for names in pruned_names:
        name = names[0]
        save_number  =  channel_savenumber_dict[name]
        weight = weights[name + '.weight'] #residual connecation choose the first layer// stride=2 layer
        scores = []
        if  'transpose' in name  or 'Transpose' in  name  : 
            scores = criterion( weight , 1 )
        else:
            scores = criterion( weight , 0 )
        scores_ = scores.copy()
        scores_.sort(reverse=True)
        if save_number == len(scores):
            threshold = scores_[save_number-1]-1
        else:
            threshold = scores_[save_number]
        pruning_mask[name]  = (torch.tensor(scores) >threshold).bool()
        if pruning_mask[name].sum() ==0:
            pruning_mask[name]  =( torch.tensor(scores) == scores[0]    ).bool()
    return pruning_mask



def pruning_model_by_mask( model , pruning_mask  , norm_layer_names, prec_layers, succ_layers):###暂时只能用BN剪枝
    """prune the model by the mask.
    Args:
        model: the torch net to be pruned.
        pruning mask :
    Returns:
        pruned model
    """
    
    WEIGHT_POSTFIX = '.weight'
    IN_CHANNEL_DIM = 1
    OUT_CHANNEL_DIM =0
    BIAS_POSTFIX = ".bias"
    MIN_CHANNELS = 2

    def group_weight_names(weight_names):
        grouped_names = {}
        for weight_name in weight_names:
            group_name = '.'.join(weight_name.split('.')[:-1])
            if group_name not in grouped_names:
                grouped_names[group_name] = [weight_name, ]
            else:
                grouped_names[group_name].append(weight_name)
        return grouped_names

    weights=model.state_dict()

    grouped_weight_names = group_weight_names(weights.keys())
 
    depthwise_mask={}
    norm_layer_names_ = norm_layer_names.copy()
 

    for id,norm_layer_name in enumerate(norm_layer_names):

        #norm_layer_name is list
        #取出mask
        prune_mask = pruning_mask[norm_layer_name[0]].byte()
        prune_indices = torch.nonzero(prune_mask).flatten()#nonzero得到数组中非零值的坐标


        for name in prec_layers[id]:
            if name in depthwise_mask.keys():
                prune_mask = depthwise_mask[name][1]
                prune_indices = depthwise_mask[name][0]
                break
        

        # 1. prune source normalization layer，先剪掉剪掉源层（BN）
        for name  in norm_layer_name:
            for weight_name in grouped_weight_names[name]:
                weights[weight_name] = weights[weight_name].masked_select(prune_mask)
        # 2. prune target succeeding conv/linear/... layers     剪掉源层（BN）后继层的（输入通道）
        for prune_layer_name in succ_layers[id]:   #遍历后继
            for weight_name in grouped_weight_names[prune_layer_name]: #遍历后继的各层(如bn层的 weights ，bias ，meaning ，var)
                if weight_name.endswith(WEIGHT_POSTFIX):#endswith以“.weight”结尾
                    transpose_flag = is_transposed(model,weight_name)
                    if (weights[weight_name].size(IN_CHANNEL_DIM) == 1):#deepwise卷积，剪枝掩码存入字典
                        depthwise_mask[prune_layer_name]=[prune_indices , prune_mask]
                        continue
                    if transpose_flag:#反卷积
                        if weights[weight_name].shape[  OUT_CHANNEL_DIM  ] != len( prune_mask  ):
                            #import pdb;pdb.set_trace()
                            factor_ = weights[weight_name].shape[  OUT_CHANNEL_DIM  ] // len( prune_mask  )
                            prune_mask_ = prune_mask.view(-1,1 )
                            prune_mask_ = prune_mask_.repeat( 1  , factor_ ).view(-1)
                            prune_indices_ = torch.nonzero(prune_mask_).flatten()
                            weights[weight_name] = weights[weight_name].index_select(OUT_CHANNEL_DIM, prune_indices_)#剪输入通道
                        else: 
                            weights[weight_name] = weights[weight_name].index_select(OUT_CHANNEL_DIM, prune_indices)#剪输入通道
                    else:#正常卷积
                        if weights[weight_name].shape[  IN_CHANNEL_DIM  ] != len( prune_mask  ):
                            #import pdb;pdb.set_trace()
                            factor_ = weights[weight_name].shape[  IN_CHANNEL_DIM  ] // len( prune_mask  )
                            prune_mask_ = prune_mask.view(-1,1 )
                            prune_mask_ = prune_mask_.repeat( 1  , factor_ ).view(-1)
                            prune_indices_ = torch.nonzero(prune_mask_).flatten()
                            weights[weight_name] = weights[weight_name].index_select(IN_CHANNEL_DIM, prune_indices_)#剪输入通道
                        else:
                            weights[weight_name] = weights[weight_name].index_select(IN_CHANNEL_DIM, prune_indices)#剪输入通道
                    
        # 3. prune target preceding conv/linear/... layers    #剪掉源层（BN）前继层的（输入通道）也就是该modules的卷积
        for prune_layer_name in prec_layers[id]: #同上
            for weight_name in grouped_weight_names[prune_layer_name]:#同上
                if weight_name.endswith(WEIGHT_POSTFIX):#endswith以“.weight”结尾
                    transpose_flag = is_transposed(model,weight_name)
                    if transpose_flag:#反卷积
                        weights[weight_name] = weights[weight_name].index_select(IN_CHANNEL_DIM, prune_indices)#剪输出通道
                    else:#正常卷积
                        weights[weight_name] = weights[weight_name].index_select(OUT_CHANNEL_DIM, prune_indices)#剪输出通道
                elif weight_name.endswith(BIAS_POSTFIX):#endswith以“.bias”结尾
                    weights[weight_name] = weights[weight_name].index_select(0, prune_indices)#剪输出通道

    pruned_model = rebuild(model, weights)
    return pruned_model
